# Copyright 2016 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


# The maximum value a signed INT type may have
DB_MAX_INT = 0x7FFFFFFF

# The cinder services binaries and topics' names
API_BINARY = "cinder-api"
SCHEDULER_BINARY = "cinder-scheduler"
VOLUME_BINARY = "cinder-volume"
BACKUP_BINARY = "cinder-backup"
SCHEDULER_TOPIC = SCHEDULER_BINARY
VOLUME_TOPIC = VOLUME_BINARY
BACKUP_TOPIC = BACKUP_BINARY
LOG_BINARIES = (SCHEDULER_BINARY, VOLUME_BINARY, BACKUP_BINARY, API_BINARY)

# The encryption key ID used by the legacy fixed-key ConfKeyMgr
FIXED_KEY_ID = '00000000-0000-0000-0000-000000000000'

# Storage protocol constants
CEPH = 'ceph'
DRBD = 'DRBD'
FC = 'FC'
FC_VARIANT_1 = 'fibre_channel'
FC_VARIANT_2 = 'fc'
FILE = 'file'
ISCSI = 'iSCSI'
ISCSI_VARIANT = 'iscsi'
ISER = 'iSER'
LIGHTOS = 'lightos'
NFS = 'NFS'
NFS_VARIANT = 'nfs'
NVMEOF = 'NVMe-oF'
NVMEOF_VARIANT_1 = 'NVMeOF'
NVMEOF_VARIANT_2 = 'nvmeof'
NVMEOF_ROCE = 'NVMe-RoCE'
NVMEOF_FC = 'NVMe-FC'
NVMEOF_TCP = 'NVMe-TCP'
SCALEIO = 'scaleio'
SCSI = 'SCSI'
STORPOOL = 'storpool'
VMDK = 'vmdk'
VSTORAGE = 'vstorageobject'

# These must be strings, because there are places that check specific type
ISCSI_VARIANTS = [ISCSI, ISCSI_VARIANT]
FC_VARIANTS = [FC, FC_VARIANT_1, FC_VARIANT_2]
NFS_VARIANTS = [NFS, NFS_VARIANT]
NVMEOF_VARIANTS = [NVMEOF, NVMEOF_VARIANT_1, NVMEOF_VARIANT_2]

CACHEABLE_PROTOCOLS = FC_VARIANTS + ISCSI_VARIANTS + NVMEOF_VARIANTS



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import socket
from unittest import mock

import ddt
from oslo_config import cfg

from cinder import exception
from cinder.tests.unit import test
from cinder.tests.unit.volume.drivers.netapp.dataontap.utils import fakes
from cinder.volume.drivers.netapp.dataontap.client import client_cmode
from cinder.volume.drivers.netapp.dataontap.client import client_cmode_rest
from cinder.volume.drivers.netapp.dataontap.utils import utils

CONF = cfg.CONF


@ddt.ddt
class NetAppCDOTDataMotionTestCase(test.TestCase):

    def setUp(self):
        super(NetAppCDOTDataMotionTestCase, self).setUp()
        self.backend = 'backend1'
        self.mock_cmode_client = self.mock_object(client_cmode, 'Client')
        self.mock_cmode_rest_client = self.mock_object(
            client_cmode_rest, 'RestClient')
        self.config = fakes.get_fake_cmode_config(self.backend)
        CONF.set_override('volume_backend_name', self.backend,
                          group=self.backend)
        CONF.set_override('netapp_transport_type', 'https',
                          group=self.backend)
        CONF.set_override('netapp_login', 'fake_user',
                          group=self.backend)
        CONF.set_override('netapp_password', 'fake_password',
                          group=self.backend)
        CONF.set_override('netapp_server_hostname', 'fake_hostname',
                          group=self.backend)
        CONF.set_override('netapp_server_port', 8866,
                          group=self.backend)
        CONF.set_override('netapp_api_trace_pattern', "fake_regex",
                          group=self.backend)
        CONF.set_override('netapp_ssl_cert_path', 'fake_ca',
                          group=self.backend)

    def test_get_backend_configuration(self):
        self.mock_object(utils, 'CONF')
        CONF.set_override('netapp_vserver', 'fake_vserver',
                          group=self.backend)
        utils.CONF.list_all_sections.return_value = [self.backend]

        config = utils.get_backend_configuration(self.backend)

        self.assertEqual('fake_vserver', config.netapp_vserver)

    def test_get_backend_configuration_different_backend_name(self):
        self.mock_object(utils, 'CONF')
        CONF.set_override('netapp_vserver', 'fake_vserver',
                          group=self.backend)
        CONF.set_override('volume_backend_name', 'fake_backend_name',
                          group=self.backend)
        utils.CONF.list_all_sections.return_value = [self.backend]

        config = utils.get_backend_configuration(self.backend)

        self.assertEqual('fake_vserver', config.netapp_vserver)
        self.assertEqual('fake_backend_name', config.volume_backend_name)

    @ddt.data([], ['fake_backend1', 'fake_backend2'])
    def test_get_backend_configuration_not_configured(self, conf_sections):
        self.mock_object(utils, 'CONF')
        utils.CONF.list_all_sections.return_value = conf_sections

        self.assertRaises(exception.ConfigNotFound,
                          utils.get_backend_configuration,
                          self.backend)

    @ddt.data(True, False)
    def test_get_client_for_backend(self, use_legacy):
        self.config.netapp_use_legacy_client = use_legacy
        self.mock_object(utils, 'get_backend_configuration',
                         return_value=self.config)

        utils.get_client_for_backend(self.backend)

        if use_legacy:
            self.mock_cmode_client.assert_called_once_with(
                hostname='fake_hostname', password='fake_password',
                username='fake_user', transport_type='https', port=8866,
                trace=mock.ANY, vserver=None, api_trace_pattern="fake_regex")
            self.mock_cmode_rest_client.assert_not_called()
        else:
            self.mock_cmode_rest_client.assert_called_once_with(
                hostname='fake_hostname', password='fake_password',
                username='fake_user', transport_type='https', port=8866,
                trace=mock.ANY, vserver=None, api_trace_pattern="fake_regex",
                ssl_cert_path='fake_ca', async_rest_timeout=60)
            self.mock_cmode_client.assert_not_called()

    @ddt.data(True, False)
    def test_get_client_for_backend_with_vserver(self, use_legacy):
        self.config.netapp_use_legacy_client = use_legacy
        self.mock_object(utils, 'get_backend_configuration',
                         return_value=self.config)

        CONF.set_override('netapp_vserver', 'fake_vserver',
                          group=self.backend)

        utils.get_client_for_backend(self.backend)

        if use_legacy:
            self.mock_cmode_client.assert_called_once_with(
                hostname='fake_hostname', password='fake_password',
                username='fake_user', transport_type='https', port=8866,
                trace=mock.ANY, vserver='fake_vserver',
                api_trace_pattern="fake_regex")
            self.mock_cmode_rest_client.assert_not_called()
        else:
            self.mock_cmode_rest_client.assert_called_once_with(
                hostname='fake_hostname', password='fake_password',
                username='fake_user', transport_type='https', port=8866,
                trace=mock.ANY, vserver='fake_vserver',
                api_trace_pattern="fake_regex", ssl_cert_path='fake_ca',
                async_rest_timeout = 60)
            self.mock_cmode_client.assert_not_called()


@ddt.ddt
class NetAppDataOntapUtilsTestCase(test.TestCase):

    def test_build_ems_log_message_0(self):

        self.mock_object(
            socket, 'gethostname', return_value='fake_hostname')

        result = utils.build_ems_log_message_0(
            'fake_driver_name', 'fake_app_version')

        expected = {
            'computer-name': 'fake_hostname',
            'event-source': 'Cinder driver fake_driver_name',
            'app-version': 'fake_app_version',
            'category': 'provisioning',
            'log-level': '5',
            'auto-support': 'false',
            'event-id': '0',
            'event-description': 'OpenStack Cinder connected to cluster node',
        }
        self.assertEqual(expected, result)

    def test_build_ems_log_message_1(self):

        self.mock_object(
            socket, 'gethostname', return_value='fake_hostname')
        aggregate_pools = ['aggr1', 'aggr2']
        flexvol_pools = ['vol1', 'vol2']

        result = utils.build_ems_log_message_1(
            'fake_driver_name', 'fake_app_version', 'fake_vserver',
            flexvol_pools, aggregate_pools)

        pool_info = {
            'pools': {
                'vserver': 'fake_vserver',
                'aggregates': aggregate_pools,
                'flexvols': flexvol_pools,
            },
        }
        self.assertDictEqual(pool_info,
                             json.loads(result['event-description']))

        result['event-description'] = ''
        expected = {
            'computer-name': 'fake_hostname',
            'event-source': 'Cinder driver fake_driver_name',
            'app-version': 'fake_app_version',
            'category': 'provisioning',
            'log-level': '5',
            'auto-support': 'false',
            'event-id': '1',
            'event-description': '',
        }
        self.assertEqual(expected, result)



#  Copyright (c) 2016 IBM Corporation
#  All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
""" Fake pyxcli exceptions for testing the driver without installing pyxcli"""


class XCLIError(Exception):
    pass


class VolumeBadNameError(XCLIError):
    pass


class CredentialsError(XCLIError):
    pass


class ConnectionError(XCLIError):
    pass


class CgHasMirrorError(XCLIError):
    pass


class CgDoesNotExistError(XCLIError):
    pass


class CgEmptyError(XCLIError):
    pass


class PoolSnapshotLimitReachedError(XCLIError):
    pass


class CommandFailedRuntimeError(XCLIError):
    pass


class PoolOutOfSpaceError(XCLIError):
    pass


class CgLimitReachedError(XCLIError):
    pass


class HostBadNameError(XCLIError):
    pass


class CgNotEmptyError(XCLIError):
    pass


class SystemOutOfSpaceError(XCLIError):
    pass


class CgNameExistsError(XCLIError):
    pass


class CgBadNameError(XCLIError):
    pass


class SnapshotGroupDoesNotExistError(XCLIError):
    pass


class ClosedTransportError(XCLIError):
    pass


class VolumeNotInConsGroup(XCLIError):
    pass



# Copyright (c) 2016 Huawei Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

from oslo_log import log as logging
from oslo_utils import strutils

from cinder.common import constants as cinder_constants
from cinder import coordination
from cinder import exception
from cinder.i18n import _
from cinder import interface
from cinder.volume import driver
from cinder.volume.drivers.huawei import common
from cinder.volume.drivers.huawei import constants
from cinder.volume.drivers.huawei import fc_zone_helper
from cinder.volume.drivers.huawei import huawei_utils
from cinder.volume.drivers.huawei import hypermetro
from cinder.zonemanager import utils as fczm_utils

LOG = logging.getLogger(__name__)


@interface.volumedriver
class HuaweiISCSIDriver(common.HuaweiBaseDriver, driver.ISCSIDriver):
    """ISCSI driver for Huawei storage arrays.

    Version history:

    .. code-block:: none

        1.0.0 - Initial driver
        1.1.0 - Provide Huawei OceanStor storage 18000 driver
        1.1.1 - Code refactor
                CHAP support
                Multiple pools support
                ISCSI multipath support
                SmartX support
                Volume migration support
                Volume retype support
        2.0.0 - Rename to HuaweiISCSIDriver
        2.0.1 - Manage/unmanage volume support
        2.0.2 - Refactor HuaweiISCSIDriver
        2.0.3 - Manage/unmanage snapshot support
        2.0.5 - Replication V2 support
        2.0.6 - Support iSCSI configuration in Replication
        2.0.7 - Hypermetro support
                Hypermetro consistency group support
                Consistency group support
                Cgsnapshot support
        2.0.8 - Backup snapshot optimal path support
        2.0.9 - Support reporting disk type of pool
    """

    VERSION = "2.0.9"

    def __init__(self, *args, **kwargs):
        super(HuaweiISCSIDriver, self).__init__(*args, **kwargs)

    def get_volume_stats(self, refresh=False):
        """Get volume status."""
        data = self._get_volume_stats(refresh=False)
        data['storage_protocol'] = cinder_constants.ISCSI
        data['driver_version'] = self.VERSION
        return data

    @coordination.synchronized('huawei-mapping-{connector[host]}')
    def initialize_connection(self, volume, connector):
        """Map a volume to a host and return target iSCSI information."""
        lun_id, lun_type = self.get_lun_id_and_type(volume)
        initiator_name = connector['initiator']
        LOG.info(
            'initiator name: %(initiator_name)s, '
            'LUN ID: %(lun_id)s.',
            {'initiator_name': initiator_name,
             'lun_id': lun_id})

        (iscsi_iqns,
         target_ips,
         portgroup_id) = self.client.get_iscsi_params(connector)
        LOG.info('initialize_connection, iscsi_iqn: %(iscsi_iqn)s, '
                 'target_ip: %(target_ip)s, '
                 'portgroup_id: %(portgroup_id)s.',
                 {'iscsi_iqn': iscsi_iqns,
                  'target_ip': target_ips,
                  'portgroup_id': portgroup_id},)

        # Create hostgroup if not exist.
        host_id = self.client.add_host_with_check(connector['host'])

        # Add initiator to the host.
        self.client.ensure_initiator_added(initiator_name,
                                           host_id)
        hostgroup_id = self.client.add_host_to_hostgroup(host_id)

        # Mapping lungroup and hostgroup to view.
        self.client.do_mapping(lun_id, hostgroup_id,
                               host_id, portgroup_id,
                               lun_type)

        hostlun_id = self.client.get_host_lun_id(host_id, lun_id,
                                                 lun_type)

        LOG.info("initialize_connection, host lun id is: %s.",
                 hostlun_id)

        chapinfo = self.client.find_chap_info(self.client.iscsi_info,
                                              initiator_name)

        # Return iSCSI properties.
        properties = {}
        properties['target_discovered'] = False
        properties['volume_id'] = volume.id
        multipath = connector.get('multipath', False)
        hostlun_id = int(hostlun_id)
        if not multipath:
            properties['target_portal'] = ('%s:3260' % target_ips[0])
            properties['target_iqn'] = iscsi_iqns[0]
            properties['target_lun'] = hostlun_id
        else:
            properties['target_iqns'] = [iqn for iqn in iscsi_iqns]
            properties['target_portals'] = [
                '%s:3260' % ip for ip in target_ips]
            properties['target_luns'] = [hostlun_id] * len(target_ips)

        # If use CHAP, return CHAP info.
        if chapinfo:
            chap_username, chap_password = chapinfo.split(';')
            properties['auth_method'] = 'CHAP'
            properties['auth_username'] = chap_username
            properties['auth_password'] = chap_password

        LOG.info("initialize_connection success. Return data: %s.",
                 strutils.mask_password(properties))
        return {'driver_volume_type': 'iscsi', 'data': properties}

    @coordination.synchronized('huawei-mapping-{connector[host]}')
    def terminate_connection(self, volume, connector, **kwargs):
        """Delete map between a volume and a host."""
        lun_id, lun_type = self.get_lun_id_and_type(volume)
        initiator_name = connector['initiator']
        host_name = connector['host']
        lungroup_id = None

        LOG.info(
            'terminate_connection: initiator name: %(ini)s, '
            'LUN ID: %(lunid)s.',
            {'ini': initiator_name,
             'lunid': lun_id},)

        portgroup = None
        portgroup_id = None
        view_id = None
        left_lunnum = -1
        ini = self.client.iscsi_info['initiators'].get(initiator_name)
        if ini and ini.get('TargetPortGroup'):
            portgroup = ini['TargetPortGroup']

        if portgroup:
            portgroup_id = self.client.get_tgt_port_group(portgroup)
        host_id = huawei_utils.get_host_id(self.client, host_name)
        if host_id:
            mapping_view_name = constants.MAPPING_VIEW_PREFIX + host_id
            view_id = self.client.find_mapping_view(mapping_view_name)
            if view_id:
                lungroup_id = self.client.find_lungroup_from_map(view_id)

        # Remove lun from lungroup.
        if lun_id and lungroup_id:
            lungroup_ids = self.client.get_lungroupids_by_lunid(
                lun_id, lun_type)
            if lungroup_id in lungroup_ids:
                self.client.remove_lun_from_lungroup(lungroup_id,
                                                     lun_id,
                                                     lun_type)
            else:
                LOG.warning("LUN is not in lungroup. "
                            "LUN ID: %(lun_id)s. "
                            "Lungroup id: %(lungroup_id)s.",
                            {"lun_id": lun_id,
                             "lungroup_id": lungroup_id})

        # Remove portgroup from mapping view if no lun left in lungroup.
        if lungroup_id:
            left_lunnum = self.client.get_obj_count_from_lungroup(lungroup_id)

        if portgroup_id and view_id and (int(left_lunnum) <= 0):
            if self.client.is_portgroup_associated_to_view(view_id,
                                                           portgroup_id):
                self.client.delete_portgroup_mapping_view(view_id,
                                                          portgroup_id)
        if view_id and (int(left_lunnum) <= 0):
            self.client.remove_chap(initiator_name)

            if self.client.lungroup_associated(view_id, lungroup_id):
                self.client.delete_lungroup_mapping_view(view_id,
                                                         lungroup_id)
            self.client.delete_lungroup(lungroup_id)
            if self.client.is_initiator_associated_to_host(initiator_name,
                                                           host_id):
                self.client.remove_iscsi_from_host(initiator_name)
            hostgroup_name = constants.HOSTGROUP_PREFIX + host_id
            hostgroup_id = self.client.find_hostgroup(hostgroup_name)
            if hostgroup_id:
                if self.client.hostgroup_associated(view_id, hostgroup_id):
                    self.client.delete_hostgoup_mapping_view(view_id,
                                                             hostgroup_id)
                self.client.remove_host_from_hostgroup(hostgroup_id,
                                                       host_id)
                self.client.delete_hostgroup(hostgroup_id)
            self.client.remove_host(host_id)
            self.client.delete_mapping_view(view_id)


@interface.volumedriver
class HuaweiFCDriver(common.HuaweiBaseDriver, driver.FibreChannelDriver):
    """FC driver for Huawei OceanStor storage arrays.

    Version history:

    .. code-block:: none

        1.0.0 - Initial driver
        1.1.0 - Provide Huawei OceanStor 18000 storage volume driver
        1.1.1 - Code refactor
                Multiple pools support
                SmartX support
                Volume migration support
                Volume retype support
                FC zone enhancement
                Volume hypermetro support
        2.0.0 - Rename to HuaweiFCDriver
        2.0.1 - Manage/unmanage volume support
        2.0.2 - Refactor HuaweiFCDriver
        2.0.3 - Manage/unmanage snapshot support
        2.0.4 - Balanced FC port selection
        2.0.5 - Replication V2 support
        2.0.7 - Hypermetro support
                Hypermetro consistency group support
                Consistency group support
                Cgsnapshot support
        2.0.8 - Backup snapshot optimal path support
        2.0.9 - Support reporting disk type of pool
    """

    VERSION = "2.0.9"

    def __init__(self, *args, **kwargs):
        super(HuaweiFCDriver, self).__init__(*args, **kwargs)
        self.fcsan = None

    def get_volume_stats(self, refresh=False):
        """Get volume status."""
        data = self._get_volume_stats(refresh=False)
        data['storage_protocol'] = cinder_constants.FC
        data['driver_version'] = self.VERSION
        return data

    def _check_fc_links(self, wwns, online_wwns_in_host, online_free_wwns,
                        host_id):
        wwns_check = []
        for wwn in wwns:
            if wwn in online_wwns_in_host or wwn in online_free_wwns:
                wwns_check.append(wwn)
                continue
            LOG.warning("Can't add FC initiator %(wwn)s to host "
                        "%(host)s, please check if this initiator has"
                        " been added to other host or isn't present "
                        "on array.", {"wwn": wwn, "host": host_id})

            if (self.configuration.min_fc_ini_online ==
                    constants.DEFAULT_MINIMUM_FC_INITIATOR_ONLINE):
                wwns_in_host = (
                    self.client.get_host_fc_initiators(host_id))
                iqns_in_host = (
                    self.client.get_host_iscsi_initiators(host_id))
                if not (wwns_in_host or iqns_in_host or
                        self.client.is_host_associated_to_hostgroup(
                            host_id)):
                    self.client.remove_host(host_id)
                msg = _("There is an FC initiator in an invalid "
                        "state. If you want to continue to attach "
                        "volume to host, configure MinFCIniOnline "
                        "in the XML file.")
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

        if len(wwns_check) < self.configuration.min_fc_ini_online:
            msg = (_("The number of online FC initiator %(wwns)s less than"
                     " the set number: %(set)s.") %
                   {"wwns": wwns_check,
                    "set": self.configuration.min_fc_ini_online})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        return wwns_check

    @coordination.synchronized('huawei-mapping-{connector[host]}')
    def initialize_connection(self, volume, connector):
        lun_id, lun_type = self.get_lun_id_and_type(volume)
        wwns = connector['wwpns']
        LOG.info(
            'initialize_connection, initiator: %(wwpns)s,'
            ' LUN ID: %(lun_id)s.',
            {'wwpns': wwns,
             'lun_id': lun_id},)

        portg_id = None
        host_id = self.client.add_host_with_check(connector['host'])

        if not self.fcsan:
            self.fcsan = fczm_utils.create_lookup_service()

        if self.fcsan:
            # Use FC switch.
            zone_helper = fc_zone_helper.FCZoneHelper(self.fcsan, self.client)
            try:
                (tgt_port_wwns, portg_id, init_targ_map) = (
                    zone_helper.build_ini_targ_map(wwns, host_id, lun_id,
                                                   lun_type))
            except Exception as err:
                self.remove_host_with_check(host_id)
                msg = _('build_ini_targ_map fails. %s') % err
                raise exception.VolumeBackendAPIException(data=msg)

            for ini in init_targ_map:
                self.client.ensure_fc_initiator_added(ini, host_id)
        else:
            # Not use FC switch.
            online_wwns_in_host = (
                self.client.get_host_online_fc_initiators(host_id))
            online_free_wwns = self.client.get_online_free_wwns()
            fc_initiators_on_array = self.client.get_fc_initiator_on_array()
            wwns = [i for i in wwns if i in fc_initiators_on_array]
            wwns = self._check_fc_links(
                wwns, online_wwns_in_host, online_free_wwns, host_id)

            for wwn in wwns:
                if wwn in online_free_wwns:
                    self.client.add_fc_port_to_host(host_id, wwn)

            (tgt_port_wwns, init_targ_map) = (
                self.client.get_init_targ_map(wwns))

        # Add host into hostgroup.
        hostgroup_id = self.client.add_host_to_hostgroup(host_id)

        metadata = huawei_utils.get_volume_private_data(volume)
        LOG.info("initialize_connection, metadata is: %s.", metadata)
        hypermetro_lun = metadata.get('hypermetro_id') is not None

        map_info = self.client.do_mapping(lun_id, hostgroup_id,
                                          host_id, portg_id,
                                          lun_type, hypermetro_lun)
        host_lun_id = self.client.get_host_lun_id(host_id, lun_id,
                                                  lun_type)

        # Return FC properties.
        fc_info = {'driver_volume_type': 'fibre_channel',
                   'data': {'target_lun': int(host_lun_id),
                            'target_discovered': True,
                            'target_wwn': tgt_port_wwns,
                            'volume_id': volume.id,
                            'initiator_target_map': init_targ_map,
                            'map_info': map_info}, }

        # Deal with hypermetro connection.
        if hypermetro_lun:
            loc_tgt_wwn = fc_info['data']['target_wwn']
            local_ini_tgt_map = fc_info['data']['initiator_target_map']
            hyperm = hypermetro.HuaweiHyperMetro(self.client,
                                                 self.rmt_client,
                                                 self.configuration)
            rmt_fc_info = hyperm.connect_volume_fc(volume, connector)

            rmt_tgt_wwn = rmt_fc_info['data']['target_wwn']
            rmt_ini_tgt_map = rmt_fc_info['data']['initiator_target_map']
            fc_info['data']['target_wwn'] = (loc_tgt_wwn + rmt_tgt_wwn)
            wwns = connector['wwpns']
            for wwn in wwns:
                if (wwn in local_ini_tgt_map
                        and wwn in rmt_ini_tgt_map):
                    fc_info['data']['initiator_target_map'][wwn].extend(
                        rmt_ini_tgt_map[wwn])

                elif (wwn not in local_ini_tgt_map
                        and wwn in rmt_ini_tgt_map):
                    fc_info['data']['initiator_target_map'][wwn] = (
                        rmt_ini_tgt_map[wwn])
                # else, do nothing

            loc_map_info = fc_info['data']['map_info']
            rmt_map_info = rmt_fc_info['data']['map_info']
            same_host_id = self._get_same_hostid(loc_map_info,
                                                 rmt_map_info)

            self.client.change_hostlun_id(loc_map_info, same_host_id)
            hyperm.rmt_client.change_hostlun_id(rmt_map_info, same_host_id)

            fc_info['data']['target_lun'] = same_host_id
            hyperm.rmt_client.logout()

        fczm_utils.add_fc_zone(fc_info)
        LOG.info("Return FC info is: %s.", fc_info)
        return fc_info

    def _get_same_hostid(self, loc_fc_info, rmt_fc_info):
        loc_aval_luns = loc_fc_info['aval_luns']
        loc_aval_luns = json.loads(loc_aval_luns)

        rmt_aval_luns = rmt_fc_info['aval_luns']
        rmt_aval_luns = json.loads(rmt_aval_luns)
        same_host_id = None

        for i in range(1, 512):
            if i in rmt_aval_luns and i in loc_aval_luns:
                same_host_id = i
                break

        LOG.info("The same hostid is: %s.", same_host_id)
        if not same_host_id:
            msg = _("Can't find the same host id from arrays.")
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        return same_host_id

    @coordination.synchronized('huawei-mapping-{connector[host]}')
    def terminate_connection(self, volume, connector, **kwargs):
        """Delete map between a volume and a host."""
        lun_id, lun_type = self.get_lun_id_and_type(volume)
        wwns = connector['wwpns']
        host_name = connector['host']
        left_lunnum = -1
        lungroup_id = None
        view_id = None
        LOG.info('terminate_connection: wwpns: %(wwns)s, '
                 'LUN ID: %(lun_id)s.',
                 {'wwns': wwns, 'lun_id': lun_id})

        host_id = huawei_utils.get_host_id(self.client, host_name)
        if host_id:
            mapping_view_name = constants.MAPPING_VIEW_PREFIX + host_id
            view_id = self.client.find_mapping_view(mapping_view_name)
            if view_id:
                lungroup_id = self.client.find_lungroup_from_map(view_id)

        if lun_id and lungroup_id:
            lungroup_ids = self.client.get_lungroupids_by_lunid(lun_id,
                                                                lun_type)
            if lungroup_id in lungroup_ids:
                self.client.remove_lun_from_lungroup(lungroup_id,
                                                     lun_id,
                                                     lun_type)
            else:
                LOG.warning("LUN is not in lungroup. "
                            "LUN ID: %(lun_id)s. "
                            "Lungroup id: %(lungroup_id)s.",
                            {"lun_id": lun_id,
                             "lungroup_id": lungroup_id})

        else:
            LOG.warning("Can't find lun on the array.")
        if lungroup_id:
            left_lunnum = self.client.get_obj_count_from_lungroup(lungroup_id)
        if int(left_lunnum) > 0:
            fc_info = {'driver_volume_type': 'fibre_channel',
                       'data': {}}
        else:
            fc_info, portg_id = self._delete_zone_and_remove_fc_initiators(
                wwns, host_id)
            if lungroup_id:
                if view_id and self.client.lungroup_associated(
                        view_id, lungroup_id):
                    self.client.delete_lungroup_mapping_view(view_id,
                                                             lungroup_id)
                self.client.delete_lungroup(lungroup_id)
            if portg_id:
                if view_id and self.client.is_portgroup_associated_to_view(
                        view_id, portg_id):
                    self.client.delete_portgroup_mapping_view(view_id,
                                                              portg_id)
                    self.client.delete_portgroup(portg_id)

            if host_id:
                hostgroup_name = constants.HOSTGROUP_PREFIX + host_id
                hostgroup_id = self.client.find_hostgroup(hostgroup_name)
                if hostgroup_id:
                    if view_id and self.client.hostgroup_associated(
                            view_id, hostgroup_id):
                        self.client.delete_hostgoup_mapping_view(
                            view_id, hostgroup_id)
                    self.client.remove_host_from_hostgroup(
                        hostgroup_id, host_id)
                    self.client.delete_hostgroup(hostgroup_id)

                if not self.client.check_fc_initiators_exist_in_host(
                        host_id):
                    self.client.remove_host(host_id)

            if view_id:
                self.client.delete_mapping_view(view_id)

        # Deal with hypermetro connection.
        metadata = huawei_utils.get_volume_private_data(volume)
        LOG.info("Detach Volume, metadata is: %s.", metadata)

        if metadata.get('hypermetro_id'):
            hyperm = hypermetro.HuaweiHyperMetro(self.client,
                                                 self.rmt_client,
                                                 self.configuration)
            hyperm.disconnect_volume_fc(volume, connector)

        LOG.info("terminate_connection, return data is: %s.",
                 fc_info)

        # This only does something if and only if the initiator_target_map
        # exists in fc_info
        fczm_utils.remove_fc_zone(fc_info)
        return fc_info

    def _delete_zone_and_remove_fc_initiators(self, wwns, host_id):
        # Get tgt_port_wwns and init_targ_map to remove zone.
        portg_id = None
        if not self.fcsan:
            self.fcsan = fczm_utils.create_lookup_service()
        if self.fcsan:
            zone_helper = fc_zone_helper.FCZoneHelper(self.fcsan,
                                                      self.client)
            (tgt_port_wwns, portg_id, init_targ_map) = (
                zone_helper.get_init_targ_map(wwns, host_id))
        else:
            (tgt_port_wwns, init_targ_map) = (
                self.client.get_init_targ_map(wwns))

        # Remove the initiators from host if need.
        if host_id:
            fc_initiators = self.client.get_host_fc_initiators(host_id)
            for wwn in wwns:
                if wwn in fc_initiators:
                    self.client.remove_fc_from_host(wwn)

        info = {'driver_volume_type': 'fibre_channel',
                'data': {'target_wwn': tgt_port_wwns,
                         'initiator_target_map': init_targ_map}}
        return info, portg_id



# Copyright 2013 Canonical Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
""" Tests for the backup service base driver. """

from unittest import mock

from oslo_serialization import jsonutils

from cinder.backup import driver
from cinder import context
from cinder import db
from cinder import exception
from cinder import objects
from cinder.tests.unit.backup import fake_service
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import test

_backup_db_fields = ['id', 'user_id', 'project_id',
                     'volume_id', 'host', 'availability_zone',
                     'display_name', 'display_description',
                     'container', 'status', 'fail_reason',
                     'service_metadata', 'service', 'size',
                     'object_count']


class BackupBaseDriverTestCase(test.TestCase):

    def _create_volume_db_entry(self, id, size):
        vol = {'id': id, 'size': size, 'status': 'available',
               'volume_type_id': self.vt['id']}
        return db.volume_create(self.ctxt, vol)['id']

    def _create_backup_db_entry(self, backupid, volid, size,
                                userid=fake.USER_ID,
                                projectid=fake.PROJECT_ID):
        backup = {'id': backupid, 'size': size, 'volume_id': volid,
                  'user_id': userid, 'project_id': projectid}
        return db.backup_create(self.ctxt, backup)['id']

    def setUp(self):
        super(BackupBaseDriverTestCase, self).setUp()
        self.ctxt = context.get_admin_context()

        self.volume_id = fake.VOLUME_ID
        self.backup_id = fake.BACKUP_ID

        self._create_backup_db_entry(self.backup_id, self.volume_id, 1)
        self._create_volume_db_entry(self.volume_id, 1)
        self.backup = objects.Backup.get_by_id(self.ctxt, self.backup_id)
        self.driver = fake_service.FakeBackupService(self.ctxt)

    def test_get_metadata(self):
        json_metadata = self.driver.get_metadata(self.volume_id)
        metadata = jsonutils.loads(json_metadata)
        self.assertEqual(2, metadata['version'])

    def test_put_metadata(self):
        metadata = {'version': 1}
        self.driver.put_metadata(self.volume_id, jsonutils.dumps(metadata))

    def test_get_put_metadata(self):
        json_metadata = self.driver.get_metadata(self.volume_id)
        self.driver.put_metadata(self.volume_id, json_metadata)

    def test_export_record(self):
        export_record = self.driver.export_record(self.backup)
        self.assertDictEqual({}, export_record)

    def test_import_record(self):
        export_record = {'key1': 'value1'}
        self.assertIsNone(self.driver.import_record(self.backup,
                                                    export_record))


class BackupMetadataAPITestCase(test.TestCase):

    def _create_volume_db_entry(self, id, size, display_name,
                                display_description):
        vol = {'id': id, 'size': size, 'status': 'available',
               'display_name': display_name,
               'display_description': display_description,
               'volume_type_id': self.vt['id']}
        return db.volume_create(self.ctxt, vol)['id']

    def setUp(self):
        super(BackupMetadataAPITestCase, self).setUp()
        self.ctxt = context.get_admin_context()
        self.volume_id = fake.VOLUME2_ID
        self.backup_id = fake.BACKUP2_ID
        self.volume_display_name = 'vol-1'
        self.volume_display_description = 'test vol'
        self._create_volume_db_entry(self.volume_id, 1,
                                     self.volume_display_name,
                                     self.volume_display_description)
        self.bak_meta_api = driver.BackupMetadataAPI(self.ctxt)

    def _add_metadata(self, vol_meta=False, vol_glance_meta=False):
        if vol_meta:
            # Add some VolumeMetadata
            db.volume_metadata_update(self.ctxt, self.volume_id,
                                      {'fee': 'fi'}, False)
            db.volume_metadata_update(self.ctxt, self.volume_id,
                                      {'fo': 'fum'}, False)

        if vol_glance_meta:
            # Add some GlanceMetadata
            db.volume_glance_metadata_create(self.ctxt, self.volume_id,
                                             'disk_format', 'bare')
            db.volume_glance_metadata_create(self.ctxt, self.volume_id,
                                             'container_type', 'ovf')

    def test_get(self):
        # Volume won't have anything other than base by default
        meta = self.bak_meta_api.get(self.volume_id)
        s1 = set(jsonutils.loads(meta).keys())
        s2 = ['version', self.bak_meta_api.TYPE_TAG_VOL_BASE_META]
        self.assertEqual(set(), s1.symmetric_difference(s2))

        self._add_metadata(vol_glance_meta=True)

        meta = self.bak_meta_api.get(self.volume_id)
        s1 = set(jsonutils.loads(meta).keys())
        s2 = ['version', self.bak_meta_api.TYPE_TAG_VOL_BASE_META,
              self.bak_meta_api.TYPE_TAG_VOL_GLANCE_META]
        self.assertEqual(set(), s1.symmetric_difference(s2))

        self._add_metadata(vol_meta=True)

        meta = self.bak_meta_api.get(self.volume_id)
        s1 = set(jsonutils.loads(meta).keys())
        s2 = ['version', self.bak_meta_api.TYPE_TAG_VOL_BASE_META,
              self.bak_meta_api.TYPE_TAG_VOL_GLANCE_META,
              self.bak_meta_api.TYPE_TAG_VOL_META]
        self.assertEqual(set(), s1.symmetric_difference(s2))

    def test_put(self):
        meta = self.bak_meta_api.get(self.volume_id)
        self.bak_meta_api.put(self.volume_id, meta)

        self._add_metadata(vol_glance_meta=True)
        meta = self.bak_meta_api.get(self.volume_id)
        self.bak_meta_api.put(self.volume_id, meta)

        self._add_metadata(vol_meta=True)
        meta = self.bak_meta_api.get(self.volume_id)
        self.bak_meta_api.put(self.volume_id, meta)

    def test_put_invalid_version(self):
        container = jsonutils.dumps({'version': 3})
        self.assertRaises(exception.BackupMetadataUnsupportedVersion,
                          self.bak_meta_api.put, self.volume_id, container)

    def test_v1_restore_factory(self):
        fact = self.bak_meta_api._v1_restore_factory()

        keys = [self.bak_meta_api.TYPE_TAG_VOL_BASE_META,
                self.bak_meta_api.TYPE_TAG_VOL_META,
                self.bak_meta_api.TYPE_TAG_VOL_GLANCE_META]

        self.assertEqual(set([]),
                         set(keys).symmetric_difference(set(fact.keys())))

        meta_container = {self.bak_meta_api.TYPE_TAG_VOL_BASE_META:
                          {'display_name': 'my-backed-up-volume',
                           'display_description': 'backed up description'},
                          self.bak_meta_api.TYPE_TAG_VOL_META: {},
                          self.bak_meta_api.TYPE_TAG_VOL_GLANCE_META: {}}

        # Emulate restore to new volume
        volume_id = fake.VOLUME3_ID
        vol_name = 'restore_backup_%s' % (self.backup_id)
        self._create_volume_db_entry(volume_id, 1, vol_name, 'fake volume')

        for f in fact:
            func = fact[f][0]
            fields = fact[f][1]
            func(meta_container[f], volume_id, fields)

        vol = db.volume_get(self.ctxt, volume_id)
        self.assertEqual('my-backed-up-volume', vol['display_name'])
        self.assertEqual('backed up description', vol['display_description'])

    def test_v1_restore_factory_no_restore_name(self):
        fact = self.bak_meta_api._v1_restore_factory()

        keys = [self.bak_meta_api.TYPE_TAG_VOL_BASE_META,
                self.bak_meta_api.TYPE_TAG_VOL_META,
                self.bak_meta_api.TYPE_TAG_VOL_GLANCE_META]

        self.assertEqual(set([]),
                         set(keys).symmetric_difference(set(fact.keys())))

        meta_container = {self.bak_meta_api.TYPE_TAG_VOL_BASE_META:
                          {'display_name': 'my-backed-up-volume',
                           'display_description': 'backed up description'},
                          self.bak_meta_api.TYPE_TAG_VOL_META: {},
                          self.bak_meta_api.TYPE_TAG_VOL_GLANCE_META: {}}
        for f in fact:
            func = fact[f][0]
            fields = fact[f][1]
            func(meta_container[f], self.volume_id, fields)

        vol = db.volume_get(self.ctxt, self.volume_id)
        self.assertEqual(self.volume_display_name, vol['display_name'])
        self.assertEqual(self.volume_display_description,
                         vol['display_description'])

    def test_v2_restore_factory(self):
        fact = self.bak_meta_api._v2_restore_factory()

        keys = [self.bak_meta_api.TYPE_TAG_VOL_BASE_META,
                self.bak_meta_api.TYPE_TAG_VOL_META,
                self.bak_meta_api.TYPE_TAG_VOL_GLANCE_META]

        self.assertEqual(set([]),
                         set(keys).symmetric_difference(set(fact.keys())))

        meta_container = {self.bak_meta_api.TYPE_TAG_VOL_BASE_META:
                          {'encryption_key_id': '123',
                           'volume_type_id': self.vt.get('id'),
                           'display_name': 'vol-2',
                           'display_description': 'description'},
                          self.bak_meta_api.TYPE_TAG_VOL_META: {},
                          self.bak_meta_api.TYPE_TAG_VOL_GLANCE_META: {}}

        for f in fact:
            func = fact[f][0]
            fields = fact[f][1]
            func(meta_container[f], self.volume_id, fields)

        vol = db.volume_get(self.ctxt, self.volume_id)
        self.assertEqual(self.volume_display_name, vol['display_name'])
        self.assertEqual(self.volume_display_description,
                         vol['display_description'])
        self.assertEqual('123', vol['encryption_key_id'])

    def test_restore_vol_glance_meta(self):
        # Fields is an empty list for _restore_vol_glance_meta method.
        fields = []
        container = {}
        self.bak_meta_api._save_vol_glance_meta(container, self.volume_id)
        self.bak_meta_api._restore_vol_glance_meta(container, self.volume_id,
                                                   fields)
        self._add_metadata(vol_glance_meta=True)
        self.bak_meta_api._save_vol_glance_meta(container, self.volume_id)
        self.bak_meta_api._restore_vol_glance_meta(container, self.volume_id,
                                                   fields)

    def test_restore_vol_meta(self):
        # Fields is an empty list for _restore_vol_meta method.
        fields = []
        container = {}
        self.bak_meta_api._save_vol_meta(container, self.volume_id)
        # Extract volume metadata from container.
        metadata = container.get('volume-metadata', {})
        self.bak_meta_api._restore_vol_meta(metadata, self.volume_id,
                                            fields)
        self._add_metadata(vol_meta=True)
        self.bak_meta_api._save_vol_meta(container, self.volume_id)
        # Extract volume metadata from container.
        metadata = container.get('volume-metadata', {})
        self.bak_meta_api._restore_vol_meta(metadata, self.volume_id, fields)

    def test_restore_vol_base_meta(self):
        # Fields is a list with 'encryption_key_id' for
        # _restore_vol_base_meta method.
        fields = ['encryption_key_id']
        container = {}
        self.bak_meta_api._save_vol_base_meta(container, self.volume_id)
        self.bak_meta_api._restore_vol_base_meta(container, self.volume_id,
                                                 fields)

    def _create_encrypted_volume_db_entry(self, id, type_id, encrypted):
        if encrypted:
            key_id = fake.ENCRYPTION_KEY_ID
            vol = {'id': id, 'size': 1, 'status': 'available',
                   'volume_type_id': type_id, 'encryption_key_id': key_id}
        else:
            vol = {'id': id, 'size': 1, 'status': 'available',
                   'volume_type_id': type_id, 'encryption_key_id': None}
        return db.volume_create(self.ctxt, vol)['id']

    def test_restore_encrypted_vol_to_different_volume_type(self):
        fields = ['encryption_key_id']
        container = {}

        # Create an encrypted volume
        enc_vol1_id = self._create_encrypted_volume_db_entry(fake.VOLUME4_ID,
                                                             'enc_vol_type',
                                                             True)

        # Create a second encrypted volume, of a different volume type
        enc_vol2_id = self._create_encrypted_volume_db_entry(fake.VOLUME5_ID,
                                                             'enc_vol_type2',
                                                             True)

        # Backup the first volume and attempt to restore to the second
        self.bak_meta_api._save_vol_base_meta(container, enc_vol1_id)
        self.assertRaises(exception.EncryptedBackupOperationFailed,
                          self.bak_meta_api._restore_vol_base_meta,
                          container[self.bak_meta_api.TYPE_TAG_VOL_BASE_META],
                          enc_vol2_id, fields)

    def test_restore_unencrypted_vol_to_different_volume_type(self):
        fields = ['encryption_key_id']
        container = {}

        # Create an unencrypted volume
        vol1_id = self._create_encrypted_volume_db_entry(fake.VOLUME6_ID,
                                                         'vol_type1',
                                                         False)

        # Create a second unencrypted volume, of a different volume type
        vol2_id = self._create_encrypted_volume_db_entry(fake.VOLUME7_ID,
                                                         'vol_type2',
                                                         False)

        # Backup the first volume and restore to the second
        self.bak_meta_api._save_vol_base_meta(container, vol1_id)
        self.bak_meta_api._restore_vol_base_meta(
            container[self.bak_meta_api.TYPE_TAG_VOL_BASE_META], vol2_id,
            fields)
        self.assertNotEqual(
            db.volume_get(self.ctxt, vol1_id)['volume_type_id'],
            db.volume_get(self.ctxt, vol2_id)['volume_type_id'])

    def test_restore_encrypted_vol_to_same_volume_type(self):
        fields = ['encryption_key_id']
        container = {}

        # Create an encrypted volume
        enc_vol1_id = self._create_encrypted_volume_db_entry(fake.VOLUME8_ID,
                                                             'enc_vol_type',
                                                             True)

        # Create an encrypted volume of the same type
        enc_vol2_id = self._create_encrypted_volume_db_entry(fake.VOLUME9_ID,
                                                             'enc_vol_type',
                                                             True)

        # Backup the first volume and restore to the second
        self.bak_meta_api._save_vol_base_meta(container, enc_vol1_id)
        self.bak_meta_api._restore_vol_base_meta(
            container[self.bak_meta_api.TYPE_TAG_VOL_BASE_META], enc_vol2_id,
            fields)

    def test_filter(self):
        metadata = {'a': 1, 'b': 2, 'c': 3}
        self.assertEqual(metadata, self.bak_meta_api._filter(metadata, []))
        self.assertEqual({'b': 2}, self.bak_meta_api._filter(metadata, ['b']))
        self.assertEqual({}, self.bak_meta_api._filter(metadata, ['d']))
        self.assertEqual({'a': 1, 'b': 2},
                         self.bak_meta_api._filter(metadata, ['a', 'b']))

    def test_save_vol_glance_meta(self):
        container = {}
        self.bak_meta_api._save_vol_glance_meta(container, self.volume_id)

    def test_save_vol_meta(self):
        container = {}
        self.bak_meta_api._save_vol_meta(container, self.volume_id)

    def test_save_vol_base_meta(self):
        container = {}
        self.bak_meta_api._save_vol_base_meta(container, self.volume_id)

    def test_is_serializable(self):
        data = {'foo': 'bar'}
        if self.bak_meta_api._is_serializable(data):
            jsonutils.dumps(data)

    def test_is_not_serializable(self):
        data = {'foo': 'bar'}
        with mock.patch.object(jsonutils, 'dumps') as mock_dumps:
            mock_dumps.side_effect = TypeError
            self.assertFalse(self.bak_meta_api._is_serializable(data))
            mock_dumps.assert_called_once_with(data)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinder.api.openstack import api_version_request
from cinder.tests.functional import api_samples_test_base


class VersionsSampleJsonTest(api_samples_test_base.ApiSampleTestBase):
    sample_dir = "versions"

    def setUp(self):
        super(VersionsSampleJsonTest, self).setUp()
        self.subs = {
            'max_api_version': api_version_request._MAX_API_VERSION}

    def test_versions_get_all(self):
        response = self.api.api_request('', strip_version=True)
        self._verify_response('versions-response',
                              self.subs,
                              response, 300, update_links=False)

    def test_versions_get_v3(self):
        response = self.api.api_request('v3/', strip_version=True)
        self._verify_response('version-show-response',
                              self.subs,
                              response, 200, update_links=False)



#    Copyright 2016 EMC Corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import fields

from cinder import exception
from cinder.i18n import _
from cinder import objects
from cinder.objects import base
from cinder.volume import group_types


@base.CinderObjectRegistry.register
class GroupType(base.CinderPersistentObject, base.CinderObject,
                base.CinderObjectDictCompat, base.CinderComparableObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    OPTIONAL_FIELDS = ['group_specs', 'projects']

    # NOTE: When adding a field obj_make_compatible needs to be updated
    fields = {
        'id': fields.UUIDField(),
        'name': fields.StringField(nullable=True),
        'description': fields.StringField(nullable=True),
        'is_public': fields.BooleanField(default=True, nullable=True),
        'projects': fields.ListOfStringsField(nullable=True),
        'group_specs': fields.DictOfNullableStringsField(nullable=True),
    }

    @classmethod
    def _get_expected_attrs(cls, context):
        return 'group_specs', 'projects'

    @classmethod
    def _from_db_object(cls, context, type, db_type, expected_attrs=None):
        if expected_attrs is None:
            expected_attrs = []
        for name, field in type.fields.items():
            if name in cls.OPTIONAL_FIELDS:
                continue
            value = db_type[name]
            if isinstance(field, fields.IntegerField):
                value = value or 0
            type[name] = value

        # Get data from db_type object that was queried by joined query
        # from DB
        if 'group_specs' in expected_attrs:
            type.group_specs = {}
            specs = db_type.get('group_specs')
            if specs and isinstance(specs, list):
                type.group_specs = {item['key']: item['value']
                                    for item in specs}
            elif specs and isinstance(specs, dict):
                type.group_specs = specs
        if 'projects' in expected_attrs:
            type.projects = db_type.get('projects', [])

        type._context = context
        type.obj_reset_changes()
        return type

    def create(self):
        if self.obj_attr_is_set('id'):
            raise exception.ObjectActionError(action='create',
                                              reason=_('already created'))
        db_group_type = group_types.create(self._context, self.name,
                                           self.group_specs,
                                           self.is_public, self.projects,
                                           self.description)
        self._from_db_object(self._context, self, db_group_type)

    def save(self):
        updates = self.cinder_obj_get_changes()
        if updates:
            group_types.update(self._context, self.id, self.name,
                               self.description)
            self.obj_reset_changes()

    def destroy(self):
        with self.obj_as_admin():
            group_types.destroy(self._context, self.id)


@base.CinderObjectRegistry.register
class GroupTypeList(base.ObjectListBase, base.CinderObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'objects': fields.ListOfObjectsField('GroupType'),
    }

    @classmethod
    def get_all(cls, context, inactive=0, filters=None, marker=None,
                limit=None, sort_keys=None, sort_dirs=None, offset=None):
        types = group_types.get_all_group_types(context, inactive, filters,
                                                marker=marker, limit=limit,
                                                sort_keys=sort_keys,
                                                sort_dirs=sort_dirs,
                                                offset=offset)
        expected_attrs = GroupType._get_expected_attrs(context)
        return base.obj_make_list(context, cls(context),
                                  objects.GroupType, types.values(),
                                  expected_attrs=expected_attrs)



# Copyright 2016 Dell Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""
Core fibre channel zone manager driver interface.

All fczm drivers should support this interface as a bare minimum.
"""

from cinder.interface import base


class FibreChannelZoneManagerDriver(base.CinderInterface):
    """FCZM driver required interface."""

    def add_connection(self, fabric, initiator_target_map, host_name=None,
                       storage_system=None):
        """Add a new initiator<>target connection.

        All implementing drivers should provide concrete implementation
        for this API.

        :param fabric: Fabric name from cinder.conf file
        :param initiator_target_map: Mapping of initiator to list of targets

        .. code-block:: python

            Example initiator_target_map:

            {
                '10008c7cff523b01': ['20240002ac000a50', '20240002ac000a40']
            }

        Note that WWPN can be in lower or upper case and can be ':'
        separated strings.
        """

    def delete_connection(self, fabric, initiator_target_map, host_name=None,
                          storage_system=None):
        """Delete an initiator<>target connection.

        :param fabric: Fabric name from cinder.conf file
        :param initiator_target_map: Mapping of initiator to list of targets

        .. code-block:: python

            Example initiator_target_map:

            {
                '10008c7cff523b01': ['20240002ac000a50', '20240002ac000a40']
            }

        Note that WWPN can be in lower or upper case and can be ':'
        separated strings.
        """

    def get_san_context(self, target_wwn_list):
        """Get SAN context for end devices.

        :param target_wwn_list: Mapping of initiator to list of targets

        Example initiator_target_map: ['20240002ac000a50', '20240002ac000a40']
        Note that WWPN can be in lower or upper case and can be
        ':' separated strings.
        """



# Copyright (C) 2020, 2021, Hitachi, Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""Fibre channel module for Hitachi HBSD Driver."""

import os

from oslo_utils import excutils

from cinder import interface
from cinder.volume import driver
from cinder.volume.drivers.hitachi import hbsd_common as common
from cinder.volume.drivers.hitachi import hbsd_rest as rest
from cinder.volume.drivers.hitachi import hbsd_rest_fc as rest_fc
from cinder.volume.drivers.hitachi import hbsd_utils as utils
from cinder.volume import volume_utils

MSG = utils.HBSDMsg

_DRIVER_INFO = {
    'version': utils.VERSION,
    'proto': 'FC',
    'hba_id': 'wwpns',
    'hba_id_type': 'World Wide Name',
    'msg_id': {
        'target': MSG.CREATE_HOST_GROUP_FAILED,
    },
    'volume_backend_name': '%(prefix)sFC' % {
        'prefix': utils.DRIVER_PREFIX,
    },
    'volume_type': 'fibre_channel',
    'param_prefix': utils.PARAM_PREFIX,
    'vendor_name': utils.VENDOR_NAME,
    'driver_dir_name': utils.DRIVER_DIR_NAME,
    'driver_prefix': utils.DRIVER_PREFIX,
    'driver_file_prefix': utils.DRIVER_FILE_PREFIX,
    'target_prefix': utils.TARGET_PREFIX,
    'hdp_vol_attr': utils.HDP_VOL_ATTR,
    'hdt_vol_attr': utils.HDT_VOL_ATTR,
    'nvol_ldev_type': utils.NVOL_LDEV_TYPE,
    'target_iqn_suffix': utils.TARGET_IQN_SUFFIX,
    'pair_attr': utils.PAIR_ATTR,
}


@interface.volumedriver
class HBSDFCDriver(driver.FibreChannelDriver):
    """Fibre channel class for Hitachi HBSD Driver.

    Version history:

    .. code-block:: none

        1.0.0 - Initial driver.
        1.1.0 - Add manage_existing/manage_existing_get_size/unmanage methods
        2.0.0 - Major redesign of the driver. This version requires the REST
                API for communication with the storage backend.
        2.1.0 - Add Cinder generic volume groups.
        2.2.0 - Add maintenance parameters.
        2.2.1 - Make the parameters name variable for supporting OEM storages.
        2.2.2 - Add Target Port Assignment.
        2.2.3 - Add port scheduler.

    """

    VERSION = utils.VERSION

    # ThirdPartySystems wiki page
    CI_WIKI_NAME = utils.CI_WIKI_NAME

    def __init__(self, *args, **kwargs):
        """Initialize instance variables."""
        utils.output_log(MSG.DRIVER_INITIALIZATION_START,
                         driver=self.__class__.__name__,
                         version=self.get_version())
        super(HBSDFCDriver, self).__init__(*args, **kwargs)

        self.configuration.append_config_values(common.COMMON_VOLUME_OPTS)
        self.configuration.append_config_values(common.COMMON_PORT_OPTS)
        self.configuration.append_config_values(rest_fc.FC_VOLUME_OPTS)
        os.environ['LANG'] = 'C'
        self.common = self._init_common(self.configuration, kwargs.get('db'))

    def _init_common(self, conf, db):
        return rest_fc.HBSDRESTFC(conf, _DRIVER_INFO, db)

    @staticmethod
    def get_driver_options():
        additional_opts = HBSDFCDriver._get_oslo_driver_opts(
            *(common._INHERITED_VOLUME_OPTS +
              rest._REQUIRED_REST_OPTS +
              ['driver_ssl_cert_verify', 'driver_ssl_cert_path',
               'san_api_port', ]))
        return (common.COMMON_VOLUME_OPTS +
                common.COMMON_PORT_OPTS +
                rest.REST_VOLUME_OPTS +
                rest_fc.FC_VOLUME_OPTS +
                additional_opts)

    def check_for_setup_error(self):
        pass

    @volume_utils.trace
    def create_volume(self, volume):
        """Create a volume and return its properties."""
        return self.common.create_volume(volume)

    @volume_utils.trace
    def create_volume_from_snapshot(self, volume, snapshot):
        """Create a volume from a snapshot and return its properties."""
        return self.common.create_volume_from_snapshot(volume, snapshot)

    @volume_utils.trace
    def create_cloned_volume(self, volume, src_vref):
        """Create a clone of the specified volume and return its properties."""
        return self.common.create_cloned_volume(volume, src_vref)

    @volume_utils.trace
    def delete_volume(self, volume):
        """Delete the specified volume."""
        self.common.delete_volume(volume)

    @volume_utils.trace
    def create_snapshot(self, snapshot):
        """Create a snapshot from a volume and return its properties."""
        return self.common.create_snapshot(snapshot)

    @volume_utils.trace
    def delete_snapshot(self, snapshot):
        """Delete the specified snapshot."""
        self.common.delete_snapshot(snapshot)

    def local_path(self, volume):
        pass

    def _update_volume_stats(self):
        """Return properties, capabilities and current states of the driver."""
        data = self.common.update_volume_stats()
        if 'pools' in data:
            data["pools"][0]["filter_function"] = self.get_filter_function()
            data["pools"][0]["goodness_function"] = (
                self.get_goodness_function())
        self._stats = data

    @volume_utils.trace
    def update_migrated_volume(
            self, ctxt, volume, new_volume, original_volume_status):
        """Do any remaining jobs after migration."""
        self.common.discard_zero_page(new_volume)
        super(HBSDFCDriver, self).update_migrated_volume(
            ctxt, volume, new_volume, original_volume_status)

    @volume_utils.trace
    def copy_image_to_volume(self, context, volume, image_service, image_id):
        """Fetch the image from image_service and write it to the volume."""
        super(HBSDFCDriver, self).copy_image_to_volume(
            context, volume, image_service, image_id)
        self.common.discard_zero_page(volume)

    @volume_utils.trace
    def extend_volume(self, volume, new_size):
        """Extend the specified volume to the specified size."""
        self.common.extend_volume(volume, new_size)

    @volume_utils.trace
    def manage_existing(self, volume, existing_ref):
        """Return volume properties which Cinder needs to manage the volume."""
        return self.common.manage_existing(volume, existing_ref)

    @volume_utils.trace
    def manage_existing_get_size(self, volume, existing_ref):
        """Return the size[GB] of the specified volume."""
        return self.common.manage_existing_get_size(existing_ref)

    @volume_utils.trace
    def unmanage(self, volume):
        """Prepare the volume for removing it from Cinder management."""
        self.common.unmanage(volume)

    @volume_utils.trace
    def do_setup(self, context):
        """Prepare for the startup of the driver."""
        self.common.do_setup(context)

    def ensure_export(self, context, volume):
        """Synchronously recreate an export for a volume."""
        pass

    def create_export(self, context, volume, connector):
        """Export the volume."""
        pass

    def remove_export(self, context, volume):
        """Remove an export for a volume."""
        pass

    def create_export_snapshot(self, context, snapshot, connector):
        pass

    def remove_export_snapshot(self, context, snapshot):
        pass

    @volume_utils.trace
    def initialize_connection(self, volume, connector):
        """Initialize connection between the server and the volume."""
        return self.common.initialize_connection(volume, connector)

    @volume_utils.trace
    def terminate_connection(self, volume, connector, **kwargs):
        """Terminate connection between the server and the volume."""
        if connector is None:
            connector = {}
        if utils.is_shared_connection(volume, connector):
            return
        self.common.terminate_connection(volume, connector)

    @volume_utils.trace
    def initialize_connection_snapshot(self, snapshot, connector, **kwargs):
        """Initialize connection between the server and the snapshot."""
        return self.common.initialize_connection(
            snapshot, connector, is_snapshot=True)

    @volume_utils.trace
    def terminate_connection_snapshot(self, snapshot, connector, **kwargs):
        """Terminate connection between the server and the snapshot."""
        self.common.terminate_connection(snapshot, connector)

    @volume_utils.trace
    def unmanage_snapshot(self, snapshot):
        """Prepare the snapshot for removing it from Cinder management."""
        return self.common.unmanage_snapshot(snapshot)

    @volume_utils.trace
    def retype(self, ctxt, volume, new_type, diff, host):
        """Retype the specified volume."""
        return self.common.retype()

    def backup_use_temp_snapshot(self):
        return True

    @volume_utils.trace
    def revert_to_snapshot(self, context, volume, snapshot):
        """Rollback the specified snapshot"""
        return self.common.revert_to_snapshot(volume, snapshot)

    @volume_utils.trace
    def create_group(self, context, group):
        return self.common.create_group()

    @volume_utils.trace
    def delete_group(self, context, group, volumes):
        return self.common.delete_group(group, volumes)

    @volume_utils.trace
    def create_group_from_src(
            self, context, group, volumes, group_snapshot=None, snapshots=None,
            source_group=None, source_vols=None):
        return self.common.create_group_from_src(
            context, group, volumes, snapshots, source_vols)

    @volume_utils.trace
    def update_group(
            self, context, group, add_volumes=None, remove_volumes=None):
        try:
            return self.common.update_group(group, add_volumes)
        except Exception:
            with excutils.save_and_reraise_exception():
                for remove_volume in remove_volumes:
                    utils.cleanup_cg_in_volume(remove_volume)

    @volume_utils.trace
    def create_group_snapshot(self, context, group_snapshot, snapshots):
        return self.common.create_group_snapshot(
            context, group_snapshot, snapshots)

    @volume_utils.trace
    def delete_group_snapshot(self, context, group_snapshot, snapshots):
        return self.common.delete_group_snapshot(group_snapshot, snapshots)



# Copyright 2021 Red Hat, Inc.
# All Rights Reserved.
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt

from cinder.api.contrib import extended_snapshot_attributes as snapshot_attr
from cinder.api import microversions as mv
from cinder.api.v3 import snapshots
from cinder import exception
from cinder.policies import snapshots as policy
from cinder.tests.unit.api import fakes as fake_api
from cinder.tests.unit.policies import base
from cinder.tests.unit import utils as test_utils


@ddt.ddt
class SnapshotsPolicyTest(base.BasePolicyTest):
    authorized_readers = [
        'legacy_admin',
        'legacy_owner',
        'system_admin',
        'project_admin',
        'project_member',
        'project_reader',
        'project_foo',
    ]

    unauthorized_readers = [
        'system_member',
        'system_reader',
        'system_foo',
        'other_project_member',
        'other_project_reader',
    ]

    authorized_members = [
        'legacy_admin',
        'legacy_owner',
        'system_admin',
        'project_admin',
        'project_member',
        'project_reader',
        'project_foo',
    ]

    unauthorized_members = [
        'system_member',
        'system_reader',
        'system_foo',
        'other_project_member',
        'other_project_reader',
    ]

    authorized_admins = [
        'legacy_admin',
        'system_admin',
        'project_admin',
    ]

    unauthorized_admins = [
        'legacy_owner',
        'system_member',
        'system_reader',
        'system_foo',
        'project_member',
        'project_reader',
        'project_foo',
        'other_project_member',
        'other_project_reader',
    ]

    # DB validations will throw SnapshotNotFound for some contexts
    unauthorized_exceptions = [
        exception.SnapshotNotFound,
    ]

    # Basic policy test is without enforcing scope (which cinder doesn't
    # yet support) and deprecated rules enabled.
    def setUp(self, enforce_scope=False, enforce_new_defaults=False,
              *args, **kwargs):
        super().setUp(enforce_scope, enforce_new_defaults, *args, **kwargs)
        self.controller = snapshots.SnapshotsController()
        self.api_path = '/v3/%s/snapshots' % (self.project_id)
        self.api_version = mv.BASE_VERSION
        self.vol_type = test_utils.create_volume_type(
            self.project_admin_context,
            name='fake_vol_type', testcase_instance=self)

    def _create_volume(self, **kwargs):
        volume = test_utils.create_volume(self.project_member_context,
                                          volume_type_id=self.vol_type.id,
                                          testcase_instance=self, **kwargs)
        return volume

    def _create_snapshot(self, **kwargs):
        volume = self._create_volume(**kwargs)
        snapshot = test_utils.create_snapshot(self.project_member_context,
                                              volume_id=volume.id,
                                              testcase_instance=self, **kwargs)
        return snapshot

    @ddt.data(*base.all_users)
    def test_get_all_policy(self, user_id):
        self._create_snapshot()
        rule_name = policy.GET_ALL_POLICY
        url = self.api_path
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)

        # Generally, any logged in user can list all volumes.
        authorized_readers = [user_id]
        unauthorized_readers = []

        # The exception is when deprecated rules are disabled, in which case
        # roles are enforced. Users without the 'reader' role should be
        # blocked.
        if self.enforce_new_defaults:
            context = self.create_context(user_id)
            if 'reader' not in context.roles:
                authorized_readers = []
                unauthorized_readers = [user_id]

        response = self.common_policy_check(user_id, authorized_readers,
                                            unauthorized_readers,
                                            self.unauthorized_exceptions,
                                            rule_name, self.controller.index,
                                            req)

        # For some users, even if they're authorized, the list of snapshots
        # will be empty if they are not in the snapshots's project.
        empty_response_users = [
            *self.unauthorized_readers,
            # legacy_admin and system_admin do not have a project_id, and
            # so the list of snapshots returned will be empty.
            'legacy_admin',
            'system_admin',
        ]
        snapshots = response['snapshots'] if response else []
        snapshot_count = 0 if user_id in empty_response_users else 1
        self.assertEqual(snapshot_count, len(snapshots))

    @ddt.data(*base.all_users)
    def test_get_policy(self, user_id):
        snapshot = self._create_snapshot()
        rule_name = policy.GET_POLICY
        url = '%s/%s' % (self.api_path, snapshot.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)

        self.common_policy_check(user_id, self.authorized_readers,
                                 self.unauthorized_readers,
                                 self.unauthorized_exceptions,
                                 rule_name, self.controller.show, req,
                                 id=snapshot.id)

    @ddt.data(*base.all_users)
    def test_extend_attribute_policy(self, user_id):
        snapshot = self._create_snapshot()
        rule_name = policy.EXTEND_ATTRIBUTE
        url = '%s/%s' % (self.api_path, snapshot.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)

        snapshot_dict = snapshot.obj_to_primitive()['versioned_object.data']
        req.get_db_snapshot = mock.MagicMock()
        req.get_db_snapshot.return_value = snapshot_dict
        resp_obj = mock.MagicMock(obj={'snapshot': snapshot_dict})
        self.assertNotIn('os-extended-snapshot-attributes:project_id',
                         snapshot_dict.keys())

        controller = snapshot_attr.ExtendedSnapshotAttributesController()

        self.common_policy_check(user_id, self.authorized_readers,
                                 self.unauthorized_readers,
                                 self.unauthorized_exceptions,
                                 rule_name, controller.show, req,
                                 resp_obj=resp_obj,
                                 id=snapshot.id, fatal=False)

        if user_id in self.authorized_readers:
            self.assertIn('os-extended-snapshot-attributes:project_id',
                          snapshot_dict.keys())

    @ddt.data(*base.all_users)
    def test_create_policy(self, user_id):
        volume = self._create_volume()
        rule_name = policy.CREATE_POLICY
        url = self.api_path
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "snapshot": {
                "name": "snap-001",
                "volume_id": volume.id,
            }
        }

        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]

        self.common_policy_check(user_id, self.authorized_members,
                                 self.unauthorized_members,
                                 unauthorized_exceptions,
                                 rule_name, self.controller.create, req,
                                 body=body)

    @ddt.data(*base.all_users)
    def test_update_policy(self, user_id):
        snapshot = self._create_snapshot()
        rule_name = policy.UPDATE_POLICY
        url = '%s/%s' % (self.api_path, snapshot.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'PUT'
        body = {
            "snapshot": {
                "description": "This is yet another snapshot."
            }
        }

        # Relax the GET_POLICY in order to get past that check.
        self.policy.set_rules({policy.GET_POLICY: ""},
                              overwrite=False)

        self.common_policy_check(user_id, self.authorized_members,
                                 self.unauthorized_members,
                                 self.unauthorized_exceptions,
                                 rule_name, self.controller.update, req,
                                 id=snapshot.id, body=body)

    @ddt.data(*base.all_users)
    def test_delete_policy(self, user_id):
        snapshot = self._create_snapshot(status='available')
        rule_name = policy.DELETE_POLICY
        url = '%s/%s' % (self.api_path, snapshot.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'DELETE'

        # Relax the GET_POLICY in order to get past that check.
        self.policy.set_rules({policy.GET_POLICY: ""},
                              overwrite=False)

        self.common_policy_check(user_id, self.authorized_members,
                                 self.unauthorized_members,
                                 self.unauthorized_exceptions,
                                 rule_name, self.controller.delete, req,
                                 id=snapshot.id)


class SnapshotsPolicySecureRbacTest(SnapshotsPolicyTest):
    authorized_readers = [
        'legacy_admin',
        'system_admin',
        'project_admin',
        'project_member',
        'project_reader',
    ]

    unauthorized_readers = [
        'legacy_owner',
        'system_member',
        'system_reader',
        'system_foo',
        'project_foo',
        'other_project_member',
        'other_project_reader',
    ]

    authorized_members = [
        'legacy_admin',
        'system_admin',
        'project_admin',
        'project_member',
    ]

    unauthorized_members = [
        'legacy_owner',
        'system_member',
        'system_reader',
        'system_foo',
        'project_reader',
        'project_foo',
        'other_project_member',
        'other_project_reader',
    ]

    def setUp(self, *args, **kwargs):
        # Test secure RBAC by disabling deprecated policy rules (scope
        # is still not enabled).
        super().setUp(enforce_scope=False, enforce_new_defaults=True,
                      *args, **kwargs)



# Copyright (c) 2016 Clinton Knight
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Performance metrics functions and cache for NetApp systems.
"""

from oslo_log import log as logging

from cinder import exception
from cinder.i18n import _


LOG = logging.getLogger(__name__)
DEFAULT_UTILIZATION = 50


class PerformanceLibrary(object):

    def __init__(self, zapi_client):

        self.zapi_client = zapi_client
        self._init_counter_info()

    def _init_counter_info(self):
        """Set a few counter names based on Data ONTAP version."""

        self.system_object_name = None
        self.avg_processor_busy_base_counter_name = None

    def _get_node_utilization(self, counters_t1, counters_t2, node_name):
        """Get node utilization from two sets of performance counters."""

        try:
            # Time spent in the single-threaded Kahuna domain
            kahuna_percent = self._get_kahuna_utilization(counters_t1,
                                                          counters_t2)

            # If Kahuna is using >60% of the CPU, the controller is fully busy
            if kahuna_percent > 60:
                return 100.0

            # Average CPU busyness across all processors
            avg_cpu_percent = 100.0 * self._get_average_cpu_utilization(
                counters_t1, counters_t2)

            # Total Consistency Point (CP) time
            total_cp_time_msec = self._get_total_consistency_point_time(
                counters_t1, counters_t2)

            # Time spent in CP Phase 2 (buffer flush)
            p2_flush_time_msec = self._get_consistency_point_p2_flush_time(
                counters_t1, counters_t2)

            # Wall-clock time between the two counter sets
            poll_time_msec = self._get_total_time(counters_t1,
                                                  counters_t2,
                                                  'total_cp_msecs')

            # If two polls happened in quick succession, use CPU utilization
            if total_cp_time_msec == 0 or poll_time_msec == 0:
                return max(min(100.0, avg_cpu_percent), 0)

            # Adjusted Consistency Point time
            adjusted_cp_time_msec = self._get_adjusted_consistency_point_time(
                total_cp_time_msec, p2_flush_time_msec)
            adjusted_cp_percent = (100.0 *
                                   adjusted_cp_time_msec / poll_time_msec)

            # Utilization is the greater of CPU busyness & CP time
            node_utilization = max(avg_cpu_percent, adjusted_cp_percent)
            return max(min(100.0, node_utilization), 0)

        except Exception:
            LOG.exception('Could not calculate node utilization for '
                          'node %s.', node_name)
            return DEFAULT_UTILIZATION

    def _get_kahuna_utilization(self, counters_t1, counters_t2):
        """Get time spent in the single-threaded Kahuna domain."""

        # Note(cknight): Because Kahuna is single-threaded, running only on
        # one CPU at a time, we can safely sum the Kahuna CPU usage
        # percentages across all processors in a node.
        return sum(self._get_performance_counter_average_multi_instance(
            counters_t1, counters_t2, 'domain_busy:kahuna',
            'processor_elapsed_time')) * 100.0

    def _get_average_cpu_utilization(self, counters_t1, counters_t2):
        """Get average CPU busyness across all processors."""

        return self._get_performance_counter_average(
            counters_t1, counters_t2, 'avg_processor_busy',
            self.avg_processor_busy_base_counter_name)

    def _get_total_consistency_point_time(self, counters_t1, counters_t2):
        """Get time spent in Consistency Points in msecs."""

        return float(self._get_performance_counter_delta(
            counters_t1, counters_t2, 'total_cp_msecs'))

    def _get_consistency_point_p2_flush_time(self, counters_t1, counters_t2):
        """Get time spent in CP Phase 2 (buffer flush) in msecs."""

        return float(self._get_performance_counter_delta(
            counters_t1, counters_t2, 'cp_phase_times:p2_flush'))

    def _get_total_time(self, counters_t1, counters_t2, counter_name):
        """Get wall clock time between two successive counters in msecs."""

        timestamp_t1 = float(self._find_performance_counter_timestamp(
            counters_t1, counter_name))
        timestamp_t2 = float(self._find_performance_counter_timestamp(
            counters_t2, counter_name))
        return (timestamp_t2 - timestamp_t1) * 1000.0

    def _get_adjusted_consistency_point_time(self, total_cp_time,
                                             p2_flush_time):
        """Get adjusted CP time by limiting CP phase 2 flush time to 20%."""

        return (total_cp_time - p2_flush_time) * 1.20

    def _get_performance_counter_delta(self, counters_t1, counters_t2,
                                       counter_name):
        """Calculate a delta value from two performance counters."""

        counter_t1 = int(
            self._find_performance_counter_value(counters_t1, counter_name))
        counter_t2 = int(
            self._find_performance_counter_value(counters_t2, counter_name))

        return counter_t2 - counter_t1

    def _get_performance_counter_average(self, counters_t1, counters_t2,
                                         counter_name, base_counter_name,
                                         instance_name=None):
        """Calculate an average value from two performance counters."""

        counter_t1 = float(self._find_performance_counter_value(
            counters_t1, counter_name, instance_name))
        counter_t2 = float(self._find_performance_counter_value(
            counters_t2, counter_name, instance_name))
        base_counter_t1 = float(self._find_performance_counter_value(
            counters_t1, base_counter_name, instance_name))
        base_counter_t2 = float(self._find_performance_counter_value(
            counters_t2, base_counter_name, instance_name))

        return (counter_t2 - counter_t1) / (base_counter_t2 - base_counter_t1)

    def _get_performance_counter_average_multi_instance(self, counters_t1,
                                                        counters_t2,
                                                        counter_name,
                                                        base_counter_name):
        """Calculate an average value from multiple counter instances."""

        averages = []
        instance_names = []
        for counter in counters_t1:
            if counter_name in counter:
                instance_names.append(counter['instance-name'])

        for instance_name in instance_names:
            average = self._get_performance_counter_average(
                counters_t1, counters_t2, counter_name, base_counter_name,
                instance_name)
            averages.append(average)

        return averages

    def _find_performance_counter_value(self, counters, counter_name,
                                        instance_name=None):
        """Given a counter set, return the value of a named instance."""

        for counter in counters:
            if counter_name in counter:
                if (instance_name is None
                        or counter['instance-name'] == instance_name):
                    return counter[counter_name]
        else:
            raise exception.NotFound(_('Counter %s not found') % counter_name)

    def _find_performance_counter_timestamp(self, counters, counter_name,
                                            instance_name=None):
        """Given a counter set, return the timestamp of a named instance."""

        for counter in counters:
            if counter_name in counter:
                if (instance_name is None
                        or counter['instance-name'] == instance_name):
                    return counter['timestamp']
        else:
            raise exception.NotFound(_('Counter %s not found') % counter_name)

    def _expand_performance_array(self, object_name, counter_name, counter):
        """Get array labels and expand counter data array."""

        # Get array labels for counter value
        counter_info = self.zapi_client.get_performance_counter_info(
            object_name, counter_name)

        array_labels = [counter_name + ':' + label.lower()
                        for label in counter_info['labels']]
        array_values = counter[counter_name].split(',')

        # Combine labels and values, and then mix into existing counter
        array_data = dict(zip(array_labels, array_values))
        counter.update(array_data)

    def _get_base_counter_name(self, object_name, counter_name):
        """Get the name of the base counter for the specified counter."""

        counter_info = self.zapi_client.get_performance_counter_info(
            object_name, counter_name)
        return counter_info['base-counter']



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""API Microversion definitions.

All new microversions should have a constant added here to be used throughout
the code instead of the specific version number. Until patches land, it's
common to end up with merge conflicts with other microversion changes. Merge
conflicts will be easier to handle via the microversion constants defined here
as the version number will only need to be changed in a single location.

Actual version numbers should be used:

  * In this file
  * In cinder/api/openstack/rest_api_version_history.rst
  * In cinder/api/openstack/api_version_request.py
  * In release notes describing the new functionality
  * In updates to api-ref

Nearly all microversion changes should include changes to all of those
locations. Make sure to add relevant documentation, and make sure that
documentation includes the final version number used.
"""

from cinder.api.openstack import api_version_request as api_version
from cinder import exception


# Add new constants here for each new microversion.

BASE_VERSION = '3.0'

UPLOAD_IMAGE_PARAMS = '3.1'

VOLUME_LIST_BOOTABLE = '3.2'

MESSAGES = '3.3'

VOLUME_LIST_GLANCE_METADATA = '3.4'

MESSAGES_PAGINATION = '3.5'

CG_UPDATE_BLANK_PROPERTIES = '3.6'

CLUSTER_SUPPORT = '3.7'

MANAGE_EXISTING_LIST = '3.8'

BACKUP_UPDATE = '3.9'

VOLUME_LIST_GROUP = '3.10'

GROUP_TYPE = '3.11'

VOLUME_SUMMARY = '3.12'

GROUP_VOLUME = '3.13'

GROUP_SNAPSHOTS = '3.14'

ETAGS = '3.15'

VOLUME_MIGRATE_CLUSTER = '3.16'

MANAGE_EXISTING_CLUSTER = '3.17'

BACKUP_PROJECT = '3.18'

GROUP_SNAPSHOT_RESET_STATUS = '3.19'

GROUP_VOLUME_RESET_STATUS = '3.20'

VOLUME_DETAIL_PROVIDER_ID = '3.21'

SNAPSHOT_LIST_METADATA_FILTER = '3.22'

VOLUME_DELETE_FORCE = '3.23'

WORKERS_CLEANUP = '3.24'

GROUP_VOLUME_LIST = '3.25'

REPLICATION_CLUSTER = '3.26'

NEW_ATTACH = '3.27'

POOL_FILTER = '3.28'

GROUP_SNAPSHOT_PAGINATION = '3.29'

SNAPSHOT_SORT = '3.30'

RESOURCE_FILTER = '3.31'

LOG_LEVEL = '3.32'

RESOURCE_FILTER_CONFIG = '3.33'

LIKE_FILTER = '3.34'

POOL_TYPE_FILTER = '3.35'

VOLUME_SUMMARY_METADATA = '3.36'

BACKUP_SORT_NAME = '3.37'

GROUP_REPLICATION = '3.38'

LIMITS_ADMIN_FILTER = '3.39'

VOLUME_REVERT = '3.40'

SNAPSHOT_LIST_USER_ID = '3.41'

VOLUME_EXTEND_INUSE = '3.42'

BACKUP_METADATA = '3.43'

NEW_ATTACH_COMPLETION = '3.44'

SUPPORT_COUNT_INFO = '3.45'

SUPPORT_NOVA_IMAGE = '3.46'

VOLUME_CREATE_FROM_BACKUP = '3.47'

VOLUME_SHARED_TARGETS_AND_SERVICE_FIELDS = '3.48'

BACKEND_STATE_REPORT = '3.49'

MULTIATTACH_VOLUMES = '3.50'

BACKUP_AZ = '3.51'

SUPPORT_VOLUME_TYPE_FILTER = '3.52'

SUPPORT_VOLUME_SCHEMA_CHANGES = '3.53'

ATTACHMENT_CREATE_MODE_ARG = '3.54'

TRANSFER_WITH_SNAPSHOTS = '3.55'

BACKUP_PROJECT_USER_ID = '3.56'

TRANSFER_WITH_HISTORY = '3.57'

GROUP_GROUPSNAPSHOT_PROJECT_ID = '3.58'

SUPPORT_TRANSFER_PAGINATION = '3.59'

VOLUME_TIME_COMPARISON_FILTER = '3.60'

VOLUME_CLUSTER_NAME = '3.61'

DEFAULT_TYPE_OVERRIDES = '3.62'

VOLUME_TYPE_ID_IN_VOLUME_DETAIL = '3.63'

ENCRYPTION_KEY_ID_IN_DETAILS = '3.64'

USE_QUOTA = '3.65'

SNAPSHOT_IN_USE = '3.66'

PROJECT_ID_OPTIONAL_IN_URL = '3.67'

SUPPORT_REIMAGE_VOLUME = '3.68'

SHARED_TARGETS_TRISTATE = '3.69'

TRANSFER_ENCRYPTED_VOLUME = '3.70'


def get_mv_header(version):
    """Gets a formatted HTTP microversion header.

    :param version: The microversion needed.
    :return: A tuple containing the microversion header with the
             requested version value.
    """
    return {'OpenStack-API-Version':
            'volume %s' % version}


def get_api_version(version):
    """Gets a ``APIVersionRequest`` instance.

    :param version: The microversion needed.
    :return: The ``APIVersionRequest`` instance.
    """
    return api_version.APIVersionRequest(version)


def get_prior_version(version):
    """Gets the microversion before the given version.

    Mostly useful for testing boundaries. This gets the microversion defined
    just prior to the given version.

    :param version: The version of interest.
    :return: The version just prior to the given version.
    """
    parts = version.split('.')

    if len(parts) != 2 or parts[0] != '3':
        raise exception.InvalidInput(reason='Version %s is not a valid '
                                     'microversion format.' % version)

    minor = int(parts[1]) - 1

    if minor < 0:
        # What's your problem? Are you trying to be difficult?
        minor = 0

    return '%s.%s' % (parts[0], minor)



# Copyright (C) 2021 NEC corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""Fibre channel module for NEC Driver."""

from cinder import interface
from cinder.volume import driver
from cinder.volume.drivers.hitachi import hbsd_common
from cinder.volume.drivers.hitachi import hbsd_fc
from cinder.volume.drivers.hitachi import hbsd_rest
from cinder.volume.drivers.hitachi import hbsd_utils
from cinder.volume.drivers.nec.v import nec_v_rest as rest
from cinder.volume.drivers.nec.v import nec_v_utils as utils

MSG = hbsd_utils.HBSDMsg


@interface.volumedriver
class VStorageFCDriver(hbsd_fc.HBSDFCDriver):
    """Fibre channel class for NEC Driver.

    Version history:

    .. code-block:: none

        1.0.0 - Initial driver.

    """

    VERSION = utils.VERSION

    # ThirdPartySystems wiki page
    CI_WIKI_NAME = utils.CI_WIKI_NAME

    def __init__(self, *args, **kwargs):
        """Initialize instance variables."""
        super(VStorageFCDriver, self).__init__(*args, **kwargs)
        self.configuration.append_config_values(rest.COMMON_VOLUME_OPTS)
        self.configuration.append_config_values(rest.FC_VOLUME_OPTS)

    def _init_common(self, conf, db):
        utils.DRIVER_INFO['proto'] = 'FC'
        utils.DRIVER_INFO['hba_id'] = 'wwpns'
        utils.DRIVER_INFO['hba_id_type'] = 'World Wide Name'
        utils.DRIVER_INFO['msg_id'] = {
            'target': MSG.CREATE_HOST_GROUP_FAILED}
        utils.DRIVER_INFO['volume_backend_name'] = '%(prefix)sFC' % {
            'prefix': utils.DRIVER_PREFIX}
        utils.DRIVER_INFO['volume_type'] = 'fibre_channel'

        return rest.VStorageRESTFC(conf, utils.DRIVER_INFO, db)

    @staticmethod
    def get_driver_options():
        additional_opts = driver.BaseVD._get_oslo_driver_opts(
            *(hbsd_common._INHERITED_VOLUME_OPTS +
              hbsd_rest._REQUIRED_REST_OPTS +
              ['driver_ssl_cert_verify', 'driver_ssl_cert_path',
               'san_api_port']))
        return (rest.COMMON_VOLUME_OPTS + rest.REST_VOLUME_OPTS +
                rest.FC_VOLUME_OPTS + additional_opts)



#
# Copyright (c) 2016 NEC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import errno
import os
import re
import traceback

from lxml import etree
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import units

from cinder import context
from cinder import exception
from cinder.i18n import _
from cinder.volume import configuration
from cinder.volume.drivers.nec import cli
from cinder.volume.drivers.san import san
from cinder.volume import qos_specs
from cinder.volume import volume_types


LOG = logging.getLogger(__name__)

FLAGS = cfg.CONF

mstorage_opts = [
    cfg.IPOpt('nec_ismcli_fip',
              default=None,
              help='FIP address of M-Series Storage iSMCLI.'),
    cfg.StrOpt('nec_ismcli_user',
               default='',
               help='User name for M-Series Storage iSMCLI.'),
    cfg.StrOpt('nec_ismcli_password',
               secret=True,
               default='',
               help='Password for M-Series Storage iSMCLI.'),
    cfg.StrOpt('nec_ismcli_privkey',
               default='',
               help='Filename of RSA private key for '
                    'M-Series Storage iSMCLI.'),
    cfg.StrOpt('nec_ldset',
               default='',
               help='M-Series Storage LD Set name for Compute Node.'),
    cfg.StrOpt('nec_ldname_format',
               default='LX:%s',
               help='M-Series Storage LD name format for volumes.'),
    cfg.StrOpt('nec_backup_ldname_format',
               default='LX:%s',
               help='M-Series Storage LD name format for snapshots.'),
    cfg.StrOpt('nec_diskarray_name',
               default='',
               help='Diskarray name of M-Series Storage.'),
    cfg.StrOpt('nec_ismview_dir',
               default='/tmp/nec/cinder',
               help='Output path of iSMview file.'),
    cfg.IntOpt('nec_ssh_pool_port_number',
               default=22,
               help='Port number of ssh pool.'),
    cfg.IntOpt('nec_unpairthread_timeout',
               default=3600,
               help='Timeout value of Unpairthread.'),
    cfg.IntOpt('nec_backend_max_ld_count',
               default=1024,
               help='Maximum number of managing sessions.'),
    cfg.BoolOpt('nec_actual_free_capacity',
                default=False,
                help='Return actual free capacity.'),
    cfg.BoolOpt('nec_ismview_alloptimize',
                default=False,
                help='Use legacy iSMCLI command with optimization.'),
    cfg.ListOpt('nec_pools',
                default=[],
                help='M-Series Storage pool numbers list to be used.'),
    cfg.ListOpt('nec_backup_pools',
                default=[],
                help='M-Series Storage backup pool number to be used.'),
    cfg.BoolOpt('nec_queryconfig_view',
                default=False,
                help='Use legacy iSMCLI command.'),
    cfg.IntOpt('nec_iscsi_portals_per_cont',
               default=0,
               deprecated_for_removal=True,
               help='Max number of iSCSI portals per controller. '
                    '0 => unlimited. '
                    'This option is deprecated and may '
                    'be removed in the next release.'),
    cfg.BoolOpt('nec_auto_accesscontrol',
                default=True,
                help='Configure access control automatically.'),
    cfg.StrOpt('nec_cv_ldname_format',
               default='LX:__ControlVolume_%xh',
               help='M-Series Storage Control Volume name format.'),
]

FLAGS.register_opts(mstorage_opts, group=configuration.SHARED_CONF_GROUP)


def convert_to_id(value62):
    alnum = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
    length = len(value62)

    weight = 0
    value = 0
    index = 0
    for i in reversed(range(0, length)):
        num = alnum.find(value62[i])
        if index != 0:
            value += int(weight * (num))
        else:
            value = num
        index += 1
        weight = 62 ** index

    value = '%032x' % value

    uuid = value[0:8]
    uuid += '-'
    uuid += value[8:12]
    uuid += '-'
    uuid += value[12:16]
    uuid += '-'
    uuid += value[16:20]
    uuid += '-'
    uuid += value[20:]

    return uuid


class MStorageVolumeCommon(object):
    """M-Series Storage volume common class."""

    def do_setup(self, context):
        self._context = context

    def check_for_setup_error(self):
        fip = self._configuration.safe_get('nec_ismcli_fip')
        user = self._configuration.safe_get('nec_ismcli_user')
        pw = self._configuration.safe_get('nec_ismcli_password')
        key = self._configuration.safe_get('nec_ismcli_privkey')
        pools = self._configuration.safe_get('nec_pools')

        if fip is None or fip == '':
            raise exception.ParameterNotFound(param='nec_ismcli_fip')
        if user is None or user == '':
            raise exception.ParameterNotFound(param='nec_ismcli_user')
        if (pw is None or pw == '') and (key is None or key == ''):
            msg = _('nec_ismcli_password nor nec_ismcli_privkey')
            raise exception.ParameterNotFound(param=msg)
        if pools is None or len(pools) == 0:
            raise exception.ParameterNotFound(param='nec_pools')

    def _set_config(self, configuration, host, driver_name):
        self._configuration = configuration
        self._host = host
        self._driver_name = driver_name

        self._configuration.append_config_values(mstorage_opts)
        self._configuration.append_config_values(san.san_opts)
        self._config_group = self._configuration.config_group

        self._properties = self._set_properties()
        self._cli = self._properties['cli']

    def _create_ismview_dir(self,
                            ismview_dir,
                            diskarray_name,
                            driver_name,
                            host):
        """Create ismview directory."""
        filename = diskarray_name
        if filename == '':
            filename = driver_name + '_' + host

        ismview_path = os.path.join(ismview_dir, filename)
        LOG.debug('ismview_path=%s.', ismview_path)
        try:
            if os.path.exists(ismview_path):
                os.remove(ismview_path)
        except OSError as e:
            with excutils.save_and_reraise_exception() as ctxt:
                if e.errno == errno.ENOENT:
                    ctxt.reraise = False

        try:
            os.makedirs(ismview_dir)
        except OSError as e:
            with excutils.save_and_reraise_exception() as ctxt:
                if e.errno == errno.EEXIST:
                    ctxt.reraise = False

        return ismview_path

    def get_conf_properties(self):
        confobj = self._configuration

        pool_pools = []
        for pool in confobj.safe_get('nec_pools'):
            if pool.endswith('h'):
                pool_pools.append(int(pool[:-1], 16))
            else:
                pool_pools.append(int(pool, 10))
        pool_backup_pools = []
        for pool in confobj.safe_get('nec_backup_pools'):
            if pool.endswith('h'):
                pool_backup_pools.append(int(pool[:-1], 16))
            else:
                pool_backup_pools.append(int(pool, 10))

        return {
            'cli_fip': confobj.safe_get('nec_ismcli_fip'),
            'cli_user': confobj.safe_get('nec_ismcli_user'),
            'cli_password': confobj.safe_get('nec_ismcli_password'),
            'cli_privkey': confobj.safe_get('nec_ismcli_privkey'),
            'pool_pools': pool_pools,
            'pool_backup_pools': pool_backup_pools,
            'pool_actual_free_capacity':
                confobj.safe_get('nec_actual_free_capacity'),
            'ldset_name': confobj.safe_get('nec_ldset'),
            'ld_name_format': confobj.safe_get('nec_ldname_format'),
            'ld_backupname_format':
                confobj.safe_get('nec_backup_ldname_format'),
            'ld_backend_max_count':
                confobj.safe_get('nec_backend_max_ld_count'),
            'thread_timeout': confobj.safe_get('nec_unpairthread_timeout'),
            'ismview_dir': confobj.safe_get('nec_ismview_dir'),
            'ismview_alloptimize': confobj.safe_get('nec_ismview_alloptimize'),
            'ssh_pool_port_number':
                confobj.safe_get('nec_ssh_pool_port_number'),
            'diskarray_name': confobj.safe_get('nec_diskarray_name'),
            'queryconfig_view': confobj.safe_get('nec_queryconfig_view'),
            'portal_number': confobj.safe_get('nec_iscsi_portals_per_cont'),
            'auto_accesscontrol': confobj.safe_get('nec_auto_accesscontrol'),
            'cv_name_format': confobj.safe_get('nec_cv_ldname_format')
        }

    def _set_properties(self):
        conf_properties = self.get_conf_properties()

        ismview_path = self._create_ismview_dir(
            conf_properties['ismview_dir'],
            conf_properties['diskarray_name'],
            self._driver_name,
            self._host)

        vendor_name, _product_dict = self.get_oem_parameter()

        backend_name = self._configuration.safe_get('volume_backend_name')
        ssh_timeout = self._configuration.safe_get('ssh_conn_timeout')
        reserved_per = self._configuration.safe_get('reserved_percentage')

        conf_properties['ssh_conn_timeout'] = ssh_timeout
        conf_properties['reserved_percentage'] = reserved_per
        conf_properties['ismview_path'] = ismview_path
        conf_properties['vendor_name'] = vendor_name
        conf_properties['products'] = _product_dict
        conf_properties['backend_name'] = backend_name
        conf_properties['cli'] = cli.MStorageISMCLI(conf_properties)

        return conf_properties

    def get_oem_parameter(self):
        product = os.path.join(os.path.dirname(__file__), 'product.xml')
        try:
            with open(product, 'r') as f:
                xml = f.read()
                root = etree.fromstring(xml)
                vendor_name = root.findall('./VendorName')[0].text

                product_dict = {}
                product_map = root.findall('./ProductMap/Product')
                for s in product_map:
                    product_dict[s.attrib['Name']] = int(s.text, 10)

                return vendor_name, product_dict
        except OSError as e:
            with excutils.save_and_reraise_exception() as ctxt:
                if e.errno == errno.ENOENT:
                    ctxt.reraise = False
            raise exception.NotFound(_('%s not found.') % product)

    @staticmethod
    def get_ldname(volid, volformat):
        alnum = ('0123456789'
                 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz')
        ldname = ""
        num = int(volid.replace(("-"), ""), 16)
        while num != 0:
            ldname = alnum[num % len(alnum)] + ldname
            num = num - num % len(alnum)
            num = num // len(alnum)

        return volformat % ldname

    def get_ldset(self, ldsets):
        ldset = None
        if self._properties['ldset_name'] == '':
            nldset = len(ldsets)
            if nldset == 0:
                msg = _('Logical Disk Set could not be found.')
                raise exception.NotFound(msg)
            else:
                ldset = None
        else:
            if self._properties['ldset_name'] not in ldsets:
                msg = (_('Logical Disk Set `%s` could not be found.') %
                       self._properties['ldset_name'])
                raise exception.NotFound(msg)
            ldset = ldsets[self._properties['ldset_name']]
        return ldset

    def get_pool_capacity(self, pools, ldsets):
        pools = [pool for (pn, pool) in pools.items()
                 if len(self._properties['pool_pools']) == 0 or
                 pn in self._properties['pool_pools']]

        free_capacity_gb = 0
        total_capacity_gb = 0
        for pool in pools:
            # Convert to GB.
            tmp_total = int(pool['total'] // units.Gi)
            tmp_free = int(pool['free'] // units.Gi)

            if free_capacity_gb < tmp_free:
                total_capacity_gb = tmp_total
                free_capacity_gb = tmp_free

        return {'total_capacity_gb': total_capacity_gb,
                'free_capacity_gb': free_capacity_gb}

    def set_backend_max_ld_count(self, xml, root):
        section = root.findall('./CMD_REQUEST')[0]
        version = section.get('version').replace('Version ', '')[0:3]
        version = float(version)
        if version < 9.1:
            if 512 < self._properties['ld_backend_max_count']:
                self._properties['ld_backend_max_count'] = 512
        else:
            if 1024 < self._properties['ld_backend_max_count']:
                self._properties['ld_backend_max_count'] = 1024

    def get_diskarray_max_ld_count(self, xml, root):
        max_ld_count = 0
        for section in root.findall(
                './'
                'CMD_REQUEST/'
                'CHAPTER[@name="Disk Array"]/'
                'OBJECT[@name="Disk Array"]/'
                'SECTION[@name="Disk Array Detail Information"]'):
            unit = section.find('./UNIT[@name="Product ID"]')
            if unit is None:
                msg = (_('UNIT[@name="Product ID"] not found. '
                         'line=%(line)d out="%(out)s"') %
                       {'line': section.sourceline, 'out': xml})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            else:
                product_id = unit.text
                if product_id in self._properties['products']:
                    max_ld_count = self._properties['products'][product_id]
                else:
                    max_ld_count = 8192
                    LOG.debug('UNIT[@name="Product ID"] unknown id. '
                              'productId=%s', product_id)
                LOG.debug('UNIT[@name="Product ID"] max_ld_count=%d.',
                          max_ld_count)
        return max_ld_count

    def get_pool_config(self, xml, root):
        pools = {}
        for xmlobj in root.findall('./'
                                   'CMD_REQUEST/'
                                   'CHAPTER[@name="Pool"]/'
                                   'OBJECT[@name="Pool"]'):
            section = xmlobj.find('./SECTION[@name="Pool Detail Information"]')
            if section is None:
                msg = (_('SECTION[@name="Pool Detail Information"] '
                         'not found. line=%(line)d out="%(out)s"') %
                       {'line': xmlobj.sourceline, 'out': xml})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            unit = section.find('./UNIT[@name="Pool No.(h)"]')
            if unit is None:
                msg = (_('UNIT[@name="Pool No.(h)"] not found. '
                         'line=%(line)d out="%(out)s"') %
                       {'line': section.sourceline, 'out': xml})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            pool_num = int(unit.text, 16)
            unit = section.find('UNIT[@name="Pool Capacity"]')
            if unit is None:
                msg = (_('UNIT[@name="Pool Capacity"] not found. '
                         'line=%(line)d out="%(out)s"') %
                       {'line': section.sourceline, 'out': xml})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            total = int(unit.text, 10)
            unit = section.find('UNIT[@name="Free Pool Capacity"]')
            if unit is None:
                msg = (_('UNIT[@name="Free Pool Capacity"] not found. '
                         'line=%(line)d out="%(out)s"') %
                       {'line': section.sourceline, 'out': xml})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            free = int(unit.text, 10)
            if self._properties['pool_actual_free_capacity']:
                unit = section.find('UNIT[@name="Used Pool Capacity"]')
                if unit is None:
                    msg = (_('UNIT[@name="Used Pool Capacity"] not found. '
                             'line=%(line)d out="%(out)s"') %
                           {'line': section.sourceline, 'out': xml})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)
                used = int(unit.text, 10)
                for section in xmlobj.findall('./SECTION[@name='
                                              '"Virtual Capacity Pool '
                                              'Information"]'):
                    unit = section.find('UNIT[@name="Actual Capacity"]')
                    if unit is None:
                        msg = (_('UNIT[@name="Actual Capacity"] not found. '
                                 'line=%(line)d out="%(out)s"') %
                               {'line': section.sourceline, 'out': xml})
                        LOG.error(msg)
                        raise exception.VolumeBackendAPIException(data=msg)
                    total = int(unit.text, 10)
                    free = total - used
            pool = {'pool_num': pool_num,
                    'total': total,
                    'free': free,
                    'ld_list': []}
            pools[pool_num] = pool
        return pools

    def get_ld_config(self, xml, root, pools):
        lds = {}
        used_ldns = []
        for section in root.findall('./'
                                    'CMD_REQUEST/'
                                    'CHAPTER[@name="Logical Disk"]/'
                                    'OBJECT[@name="Logical Disk"]/'
                                    'SECTION[@name="LD Detail Information"]'):
            unit = section.find('./UNIT[@name="LDN(h)"]')
            if unit is None:
                msg = (_('UNIT[@name="LDN(h)"] not found. '
                         'line=%(line)d out="%(out)s"') %
                       {'line': section.sourceline, 'out': xml})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            ldn = int(unit.text, 16)
            unit = section.find('./UNIT[@name="OS Type"]')
            if unit is None:
                msg = (_('UNIT[@name="OS Type"] not found. '
                         'line=%(line)d out="%(out)s"') %
                       {'line': section.sourceline, 'out': xml})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            ostype = unit.text if unit.text is not None else ''
            unit = section.find('./UNIT[@name="LD Name"]')
            if unit is None:
                msg = (_('UNIT[@name="LD Name"] not found. '
                         'line=%(line)d out="%(out)s"') %
                       {'line': section.sourceline, 'out': xml})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            ldname = ostype + ':' + unit.text
            unit = section.find('./UNIT[@name="Pool No.(h)"]')
            if unit is None:
                msg = (_('UNIT[@name="Pool No.(h)"] not found. '
                         'line=%(line)d out="%(out)s"') %
                       {'line': section.sourceline, 'out': xml})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            pool_num = int(unit.text, 16)

            unit = section.find('./UNIT[@name="LD Capacity"]')
            if unit is None:
                msg = (_('UNIT[@name="LD Capacity"] not found. '
                         'line=%(line)d out="%(out)s"') %
                       {'line': section.sourceline, 'out': xml})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            # byte capacity transform GB capacity.
            ld_capacity = int(unit.text, 10) // units.Gi

            unit = section.find('./UNIT[@name="RPL Attribute"]')
            if unit is None:
                msg = (_('UNIT[@name="RPL Attribute"] not found. '
                         'line=%(line)d out="%(out)s"') %
                       {'line': section.sourceline, 'out': xml})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            rplatr = unit.text

            unit = section.find('./UNIT[@name="Purpose"]')
            if unit is None:
                msg = (_('UNIT[@name="Purpose"] not found. '
                         'line=%(line)d out="%(out)s"') %
                       {'line': section.sourceline, 'out': xml})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            purpose = unit.text

            ld = {'ldname': ldname,
                  'ldn': ldn,
                  'pool_num': pool_num,
                  'ld_capacity': ld_capacity,
                  'RPL Attribute': rplatr,
                  'Purpose': purpose}
            pools[pool_num]['ld_list'].append(ld)
            lds[ldname] = ld
            used_ldns.append(ldn)
        return lds, used_ldns

    def get_iscsi_ldset_config(self, xml, root):
        ldsets = {}
        for xmlobj in root.findall('./'
                                   'CMD_REQUEST/'
                                   'CHAPTER[@name="Access Control"]/'
                                   'OBJECT[@name="LD Set(iSCSI)"]'):
            ldsetlds = {}
            portals = []
            initiators = []
            for unit in xmlobj.findall('./SECTION[@name="Portal"]/'
                                       'UNIT[@name="Portal"]'):
                if not unit.text.startswith('0.0.0.0:'):
                    portals.append(unit.text)

            for unit in xmlobj.findall('./SECTION[@name="Initiator List"]/'
                                       'UNIT[@name="Initiator List"]'):
                initiators.append(unit.text)

            section = xmlobj.find('./SECTION[@name="LD Set(iSCSI)'
                                  ' Information"]')
            if section is None:
                return ldsets
            unit = section.find('./UNIT[@name="Platform"]')
            if unit is None:
                msg = (_('UNIT[@name="Platform"] not found. '
                         'line=%(line)d out="%(out)s"') %
                       {'line': section.sourceline, 'out': xml})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            platform = unit.text
            unit = section.find('./UNIT[@name="LD Set Name"]')
            if unit is None:
                msg = (_('UNIT[@name="LD Set Name"] not found. '
                         'line=%(line)d out="%(out)s"') %
                       {'line': section.sourceline, 'out': xml})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            ldsetname = platform + ':' + unit.text
            unit = section.find('./UNIT[@name="Target Mode"]')
            if unit is None:
                msg = (_('UNIT[@name="Target Mode"] not found. '
                         'line=%(line)d out="%(out)s"') %
                       {'line': section.sourceline, 'out': xml})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            tmode = unit.text
            if tmode == 'Normal':
                unit = section.find('./UNIT[@name="Target Name"]')
                if unit is None:
                    msg = (_('UNIT[@name="Target Name"] not found. '
                             'line=%(line)d out="%(out)s"') %
                           {'line': section.sourceline, 'out': xml})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)
                iqn = unit.text
                for section in xmlobj.findall('./SECTION[@name='
                                              '"LUN/LD List"]'):
                    unit = section.find('./UNIT[@name="LDN(h)"]')
                    if unit is None:
                        msg = (_('UNIT[@name="LDN(h)"] not found. '
                                 'line=%(line)d out="%(out)s"') %
                               {'line': section.sourceline, 'out': xml})
                        LOG.error(msg)
                        raise exception.VolumeBackendAPIException(data=msg)
                    ldn = int(unit.text, 16)
                    unit = section.find('./UNIT[@name="LUN(h)"]')
                    if unit is None:
                        msg = (_('UNIT[@name="LUN(h)"] not found. '
                                 'line=%(line)d out="%(out)s"') %
                               {'line': section.sourceline, 'out': xml})
                        LOG.error(msg)
                        raise exception.VolumeBackendAPIException(data=msg)
                    lun = int(unit.text, 16)
                    ld = {'ldn': ldn,
                          'lun': lun,
                          'iqn': iqn}
                    ldsetlds[ldn] = ld
            elif tmode == 'Multi-Target':
                for section in xmlobj.findall('./SECTION[@name='
                                              '"Target Information For '
                                              'Multi-Target Mode"]'):
                    unit = section.find('./UNIT[@name="Target Name"]')
                    if unit is None:
                        msg = (_('UNIT[@name="Target Name"] not found. '
                                 'line=%(line)d out="%(out)s"') %
                               {'line': section.sourceline, 'out': xml})
                        LOG.error(msg)
                        raise exception.VolumeBackendAPIException(data=msg)
                    iqn = unit.text
                    unit = section.find('./UNIT[@name="LDN(h)"]')
                    if unit is None:
                        msg = (_('UNIT[@name="LDN(h)"] not found. '
                                 'line=%(line)d out="%(out)s"') %
                               {'line': section.sourceline, 'out': xml})
                        LOG.error(msg)
                        raise exception.VolumeBackendAPIException(data=msg)
                    if unit.text.startswith('-'):
                        continue
                    ldn = int(unit.text, 16)
                    unit = section.find('./UNIT[@name="LUN(h)"]')
                    if unit is None:
                        msg = (_('UNIT[@name="LUN(h)"] not found. '
                                 'line=%(line)d out="%(out)s"') %
                               {'line': section.sourceline, 'out': xml})
                        LOG.error(msg)
                        raise exception.VolumeBackendAPIException(data=msg)
                    if unit.text.startswith('-'):
                        continue
                    lun = int(unit.text, 16)
                    ld = {'ldn': ldn,
                          'lun': lun,
                          'iqn': iqn}
                    ldsetlds[ldn] = ld
            else:
                LOG.debug('`%(mode)s` Unknown Target Mode. '
                          'line=%(line)d out="%(out)s"',
                          {'mode': tmode, 'line': unit.sourceline, 'out': xml})
            ldset = {'ldsetname': ldsetname,
                     'protocol': 'iSCSI',
                     'mode': tmode,
                     'portal_list': portals,
                     'lds': ldsetlds,
                     'initiator_list': initiators}
            ldsets[ldsetname] = ldset
        return ldsets

    def get_fc_ldset_config(self, xml, root):
        ldsets = {}
        for xmlobj in root.findall('./'
                                   'CMD_REQUEST/'
                                   'CHAPTER[@name="Access Control"]/'
                                   'OBJECT[@name="LD Set(FC)"]'):
            ldsetlds = {}
            section = xmlobj.find('./SECTION[@name="LD Set(FC)'
                                  ' Information"]')
            if section is None:
                return ldsets
            unit = section.find('./UNIT[@name="Platform"]')
            if unit is None:
                msg = (_('UNIT[@name="Platform"] not found. '
                         'line=%(line)d out="%(out)s"') %
                       {'line': section.sourceline, 'out': xml})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            platform = unit.text
            unit = section.find('./UNIT[@name="LD Set Name"]')
            if unit is None:
                msg = (_('UNIT[@name="LD Set Name"] not found. '
                         'line=%(line)d out="%(out)s"') %
                       {'line': section.sourceline, 'out': xml})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            ldsetname = platform + ':' + unit.text
            wwpns = []
            ports = []
            for section in xmlobj.findall('./SECTION[@name="Path List"]'):
                unit = section.find('./UNIT[@name="Path"]')
                if unit is None:
                    msg = (_('UNIT[@name="Path"] not found. '
                             'line=%(line)d out="%(out)s"') %
                           {'line': section.sourceline, 'out': xml})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)
                if unit.text.find('(') != -1:
                    ports.append(unit.text)
                else:
                    wwpns.append(unit.text)
            for section in xmlobj.findall('./SECTION[@name="LUN/LD List"]'):
                unit = section.find('./UNIT[@name="LDN(h)"]')
                if unit is None:
                    msg = (_('UNIT[@name="LDN(h)"] not found. '
                             'line=%(line)d out="%(out)s"') %
                           {'line': section.sourceline, 'out': xml})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)
                ldn = int(unit.text, 16)
                unit = section.find('./UNIT[@name="LUN(h)"]')
                if unit is None:
                    msg = (_('UNIT[@name="LUN(h)"] not found. '
                             'line=%(line)d out="%(out)s"') %
                           {'line': section.sourceline, 'out': xml})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)
                lun = int(unit.text, 16)
                ld = {'ldn': ldn,
                      'lun': lun}
                ldsetlds[ldn] = ld
            ldset = {'ldsetname': ldsetname,
                     'lds': ldsetlds,
                     'protocol': 'FC',
                     'wwpn': wwpns,
                     'port': ports}
            ldsets[ldsetname] = ldset
        return ldsets

    def get_hostport_config(self, xml, root):
        hostports = {}
        for section in root.findall('./'
                                    'CMD_REQUEST/'
                                    'CHAPTER[@name="Controller"]/'
                                    'OBJECT[@name="Host Port"]/'
                                    'SECTION[@name="Host Director'
                                    '/Host Port Information"]'):
            unit = section.find('./UNIT[@name="Port No.(h)"]')
            if unit is None:
                msg = (_('UNIT[@name="Port No.(h)"] not found. '
                         'line=%(line)d out="%(out)s"') %
                       {'line': section.sourceline, 'out': xml})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            units = unit.text.split('-')
            director = int(units[0], 16)
            port = int(units[1], 16)
            unit = section.find('./UNIT[@name="IP Address"]')
            if unit is None:
                unit = section.find('./UNIT[@name="WWPN"]')
                if unit is None:
                    msg = (_('UNIT[@name="WWPN"] not found. '
                             'line=%(line)d out="%(out)s"') %
                           {'line': section.sourceline, 'out': xml})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)
                wwpn = unit.text
                hostport = {
                    'director': director,
                    'port': port,
                    'wwpn': wwpn,
                    'protocol': 'FC',
                }
            else:
                ip = unit.text
                if ip == '0.0.0.0':
                    continue

                hostport = {
                    'director': director,
                    'port': port,
                    'ip': ip,
                    'protocol': 'iSCSI',
                }
            if director not in hostports:
                hostports[director] = []
            hostports[director].append(hostport)
        return hostports

    def configs(self, xml):
        root = etree.fromstring(xml)
        pools = self.get_pool_config(xml, root)
        lds, used_ldns = self.get_ld_config(xml, root, pools)
        iscsi_ldsets = self.get_iscsi_ldset_config(xml, root)
        fc_ldsets = self.get_fc_ldset_config(xml, root)
        hostports = self.get_hostport_config(xml, root)
        diskarray_max_ld_count = self.get_diskarray_max_ld_count(xml, root)

        self.set_backend_max_ld_count(xml, root)

        ldsets = {}
        ldsets.update(iscsi_ldsets)
        ldsets.update(fc_ldsets)

        return pools, lds, ldsets, used_ldns, hostports, diskarray_max_ld_count

    def get_xml(self):
        ismview_path = self._properties['ismview_path']
        if os.path.exists(ismview_path) and os.path.isfile(ismview_path):
            with open(ismview_path, 'r') as f:
                xml = f.read()
                LOG.debug('loaded from %s.', ismview_path)
        else:
            xml = self._cli.view_all(ismview_path, False, False)
        return xml

    def parse_xml(self):
        try:
            xml = self.get_xml()
            return self.configs(xml)
        except Exception:
            LOG.debug('parse_xml Unexpected error. exception=%s',
                      traceback.format_exc())
            xml = self._cli.view_all(self._properties['ismview_path'], False)
            return self.configs(xml)

    def get_volume_type_qos_specs(self, volume_type_id):
        specs = {}

        ctxt = context.get_admin_context()
        if volume_type_id is not None:
            volume_type = volume_types.get_volume_type(ctxt, volume_type_id)

            qos_specs_id = volume_type.get('qos_specs_id')
            if qos_specs_id is not None:
                specs = qos_specs.get_qos_specs(ctxt, qos_specs_id)['specs']

            LOG.debug('get_volume_type_qos_specs '
                      'volume_type=%(volume_type)s, '
                      'qos_specs_id=%(qos_spec_id)s, '
                      'specs=%(specs)s',
                      {'volume_type': volume_type,
                       'qos_spec_id': qos_specs_id,
                       'specs': specs})
        return specs

    def get_qos_parameters(self, specs, reset):
        qos_params = {}
        if 'upperlimit' in specs and specs['upperlimit'] is not None:
            if self.validates_number(specs['upperlimit']) is True:
                upper_limit = int(specs['upperlimit'], 10)
                if ((upper_limit != 0) and
                        ((upper_limit < 10) or (upper_limit > 1000000))):
                    raise exception.InvalidConfigurationValue(
                        value=upper_limit, option='upperlimit')
                qos_params['upperlimit'] = upper_limit
            else:
                raise exception.InvalidConfigurationValue(
                    value=specs['upperlimit'], option='upperlimit')
        else:
            # 0: Set to no limit.(default)
            #    On the QoS function in NEC Storage, 0 means there is no
            #    limit.
            # None: Keep current value.
            qos_params['upperlimit'] = 0 if reset else None

        if 'lowerlimit' in specs and specs['lowerlimit'] is not None:
            if self.validates_number(specs['lowerlimit']) is True:
                lower_limit = int(specs['lowerlimit'], 10)
                if (lower_limit != 0 and (lower_limit < 10 or
                                          lower_limit > 1000000)):
                    raise exception.InvalidConfigurationValue(
                        value=lower_limit, option='lowerlimit')
                qos_params['lowerlimit'] = lower_limit
            else:
                raise exception.InvalidConfigurationValue(
                    value=specs['lowerlimit'], option='lowerlimit')
        else:
            # 0: Set to no limit.(default)
            #    On the QoS function in NEC Storage, 0 means there is no
            #    limit.
            # None: Keep current value.
            qos_params['lowerlimit'] = 0 if reset else None

        if 'upperreport' in specs:
            if specs['upperreport'] not in ['on', 'off']:
                LOG.debug('Illegal arguments. '
                          'upperreport is not on or off.'
                          'upperreport=%s', specs['upperreport'])
                qos_params['upperreport'] = 'off' if reset else None
            else:
                qos_params['upperreport'] = specs['upperreport']
        else:
            # off: Set to no report.(default)
            # None: Keep current value.
            qos_params['upperreport'] = 'off' if reset else None
        return qos_params

    def check_accesscontrol(self, ldsets, ld):
        """Check Logical disk is in-use or not."""
        set_accesscontrol = False
        for ldset in ldsets.values():
            if ld['ldn'] in ldset['lds']:
                set_accesscontrol = True
                break
        return set_accesscontrol

    def validates_number(self, value):
        return re.match(r'^(?![-+]0+$)[-+]?([1-9][0-9]*)?[0-9](\.[0-9]+)?$',
                        '%s' % value) and True or False



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_concurrency import processutils
from oslo_log import log as logging

from cinder import exception
from cinder.i18n import _
from cinder.image import accelerator
from cinder import utils

LOG = logging.getLogger(__name__)


class AccelGZIP(accelerator.AccelBase):
    def is_accel_exist(self):
        cmd = ['which', 'gzip']
        try:
            utils.execute(*cmd)
        except processutils.ProcessExecutionError:
            LOG.error("GZIP package is not installed.")
            return False

        return True

    # NOTE(ZhengMa): Gzip compresses a file in-place and adds a .gz
    # extension to the filename, so we rename the compressed file back
    # to the name Cinder expects it to have.
    # (Cinder expects to have A to upload)
    # Follow these steps:
    # 1. compress A to A.gz (gzip_out_file is A.gz)
    # 2. mv A.gz to A (gzip_out_file to dest)
    def compress_img(self, src, dest, run_as_root):
        try:
            gzip_compress_cmd = ['gzip', '-k', src]
            utils.execute(*gzip_compress_cmd, run_as_root=run_as_root)
        except processutils.ProcessExecutionError as ex:
            raise exception.CinderAcceleratorError(
                accelerator='GZIP',
                description=_("Volume compression failed while "
                              "uploading to glance. GZIP compression "
                              "command failed."),
                cmd=gzip_compress_cmd,
                reason=ex.stderr)
        try:
            gzip_output_filename = src + '.gz'
            mv_cmd = ['mv', gzip_output_filename, dest]
            utils.execute(*mv_cmd, run_as_root=run_as_root)
        except processutils.ProcessExecutionError as ex:
            fnames = {'i_fname': gzip_output_filename, 'o_fname': dest}
            raise exception.CinderAcceleratorError(
                accelerator='GZIP',
                description = _("Failed to rename %(i_fname)s "
                                "to %(o_fname)s") % fnames,
                cmd=mv_cmd,
                reason=ex.stderr)

    # NOTE(ZhengMa): Gzip can only decompresses a file with a .gz
    # extension to the filename, so we rename the original file so
    # that it can be accepted by Gzip.
    # Follow these steps:
    # 1. mv A to A.gz (gzip_in_file is A.gz)
    # 2. decompress A.gz to A (gzip_in_file to dest)
    def decompress_img(self, src, dest, run_as_root):
        try:
            gzip_input_filename = dest + '.gz'
            mv_cmd = ['mv', src, gzip_input_filename]
            utils.execute(*mv_cmd, run_as_root=run_as_root)
        except processutils.ProcessExecutionError as ex:
            fnames = {'i_fname': src, 'o_fname': gzip_input_filename}
            raise exception.CinderAcceleratorError(
                accelerator='GZIP',
                description = _("Failed to rename %(i_fname)s "
                                "to %(o_fname)s") % fnames,
                cmd=mv_cmd,
                reason=ex.stderr)
        try:
            gzip_decompress_cmd = ['gzip', '-d', gzip_input_filename]
            utils.execute(*gzip_decompress_cmd, run_as_root=run_as_root)
        except processutils.ProcessExecutionError as ex:
            raise exception.CinderAcceleratorError(
                accelerator='GZIP',
                description = _("Image decompression failed while "
                                "downloading from glance. GZIP "
                                "decompression command failed."),
                cmd=gzip_decompress_cmd,
                reason=ex.stderr)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


def stub_copy_volume_to_image(self, context, volume, metadata, force):
    image_metadata = {
        "status": "uploading",
        "container_format": "bare",
        "image_name": "test",
        "visibility": "private",
        "updated_at": "2017-06-05T08:44:28.000000",
        "image_id": "de75b74e-7f0d-4b59-a263-bd87bfc313bd",
        "display_description": None,
        "id": "3a81fdac-e8ae-4e61-b6a2-2e14ff316f19",
        "size": 1,
        "disk_format": "raw",
        "volume_type": None,
        "protected": False
    }
    return image_metadata


def stub_manage_existing(self, req, body):
    volume = {
        "volume": {
            "status": "creating",
            "user_id": "eae1472b5fc5496998a3d06550929e7e",
            "attachments": [],
            "links": [
                {
                    "href":
                        "http://10.0.2.15:8776/v3/87c8522052ca4eed98bc672b4c1a"
                        "3ddb/volumes/23cf872b-c781-4cd4-847d-5f2ec8cbd91c",
                    "rel": "self"
                },
                {
                    "href": "http://10.0.2.15:8776/87c8522052ca4eed98bc672b4c1"
                            "a3ddb/volumes/23cf872b-c781-4cd4-847d-5f2ec8cbd91"
                            "c",
                    "rel": "bookmark"
                }
            ],
            "availability_zone": "az2",
            "bootable": "false",
            "encrypted": "false",
            "created_at": "2014-07-18T00:12:54.000000",
            "description": "Volume imported from existingLV",
            "os-vol-tenant-attr:tenant_id": "87c8522052ca4eed98bc672b4c1a3ddb",
            "volume_type": "null",
            "name": "New Volume",
            "source_volid": "null",
            "snapshot_id": "null",
            "metadata": {
                "key2": "value2",
                "key1": "value1"
            },
            "id": "23cf872b-c781-4cd4-847d-5f2ec8cbd91c",
            "size": 0
        }
    }

    return volume


def stub_manage_existing_snapshot(self, req, body):
    snapshot = {
        "snapshot": {
            "status": "creating",
            "size": 1,
            "metadata": {
                "manage-snap-meta1": "value1",
                "manage-snap-meta3": "value3",
                "manage-snap-meta2": "value2"
            },
            "name": "new_snapshot",
            "volume_id": "1df34919-aba7-4a1b-a614-3b409d71ac03",
            "created_at": "2018-09-26T03:45:03.893592",
            "description": "this is a new snapshot",
            "id": "b6314a71-9d3d-439a-861d-b790def0d693",
            "updated_at": "null"
        }
    }

    return snapshot



# Copyright 2012 Intel Inc, OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Fakes For filters tests.
"""


class FakeHostManager(object):
    """Defines fake hosts.

    host1: free_ram_mb=1024-512-512=0, free_disk_gb=1024-512-512=0
    host2: free_ram_mb=2048-512=1536  free_disk_gb=2048-512=1536
    host3: free_ram_mb=4096-1024=3072  free_disk_gb=4096-1024=3072
    host4: free_ram_mb=8192  free_disk_gb=8192
    """

    def __init__(self):
        self.service_states = {
            'host1': {
                'compute': {'host_memory_free': 1073741824},
            },
            'host2': {
                'compute': {'host_memory_free': 2147483648},
            },
            'host3': {
                'compute': {'host_memory_free': 3221225472},
            },
            'host4': {
                'compute': {'host_memory_free': 999999999},
            },
        }


class FakeHostState(object):
    def __init__(self, host, attribute_dict):
        self.host = host
        for (key, val) in attribute_dict.items():
            setattr(self, key, val)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinder.tests.functional import api_samples_test_base as test_base


class LimitsSampleJsonTest(test_base.VolumesSampleBase):
    sample_dir = "limits"

    def setUp(self):
        super(LimitsSampleJsonTest, self).setUp()

    def test_limits_get(self):
        response = self._do_get('limits')
        self._verify_response('limits-show-response', {}, response, 200)
